"""Module state store tests.

DEPRECATED: Testing ModuleStore independently of ModuleView is no longer helpful.
Try to add new tests to test_module_state.py, where they can be tested together,
treating ModuleState as a private implementation detail.
"""

from typing import Dict, List, Optional, Set, cast

import pytest
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.robot.types import RobotType

from opentrons.hardware_control.modules.types import LiveData
from opentrons.protocol_engine import actions, commands
from opentrons.protocol_engine.commands import (
    heater_shaker as hs_commands,
)
from opentrons.protocol_engine.commands import (
    temperature_module as temp_commands,
)
from opentrons.protocol_engine.commands import (
    thermocycler as tc_commands,
)
from opentrons.protocol_engine.resources import deck_configuration_provider
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.addressable_areas import (
    AddressableAreaState,
    AddressableAreaView,
)
from opentrons.protocol_engine.state.config import Config
from opentrons.protocol_engine.state.module_substates import (
    AbsorbanceReaderId,
    AbsorbanceReaderSubState,
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
    MagneticModuleId,
    MagneticModuleSubState,
    ModuleSubStateType,
    TemperatureModuleId,
    TemperatureModuleSubState,
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
)
from opentrons.protocol_engine.state.modules import (
    HardwareModule,
    ModuleState,
    ModuleStore,
)
from opentrons.protocol_engine.types import (
    AddressableArea,
    DeckConfigurationType,
    DeckSlotLocation,
    DeckType,
    HeaterShakerLatchStatus,
    ModuleDefinition,
    ModuleModel,
    PotentialCutoutFixture,
)
from opentrons.types import DeckSlotName

_OT2_STANDARD_CONFIG = Config(
    use_simulated_deck_config=False,
    robot_type="OT-2 Standard",
    deck_type=DeckType.OT2_STANDARD,
)


def get_addressable_area_view(
    loaded_addressable_areas_by_name: Optional[Dict[str, AddressableArea]] = None,
    potential_cutout_fixtures_by_cutout_id: Optional[
        Dict[str, Set[PotentialCutoutFixture]]
    ] = None,
    deck_definition: Optional[DeckDefinitionV5] = None,
    deck_configuration: Optional[DeckConfigurationType] = None,
    robot_type: RobotType = "OT-3 Standard",
    use_simulated_deck_config: bool = False,
) -> AddressableAreaView:
    """Get a labware view test subject."""
    state = AddressableAreaState(
        loaded_addressable_areas_by_name=loaded_addressable_areas_by_name or {},
        potential_cutout_fixtures_by_cutout_id=potential_cutout_fixtures_by_cutout_id
        or {},
        deck_definition=deck_definition or cast(DeckDefinitionV5, {"otId": "fake"}),
        deck_configuration=deck_configuration or [],
        robot_definition={
            "displayName": "OT-3",
            "robotType": "OT-3 Standard",
            "models": ["OT-3 Standard"],
            "extents": [477.2, 493.8, 0.0],
            "paddingOffsets": {
                "rear": -177.42,
                "front": 51.8,
                "leftSide": 31.88,
                "rightSide": -80.32,
            },
            "mountOffsets": {
                "left": [-13.5, -60.5, 255.675],
                "right": [40.5, -60.5, 255.675],
                "gripper": [84.55, -12.75, 93.85],
            },
        },
        robot_type=robot_type,
        use_simulated_deck_config=use_simulated_deck_config,
    )

    return AddressableAreaView(state=state)


def test_initial_state() -> None:
    """It should initialize the module state."""
    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    assert subject.state == ModuleState(
        deck_type=DeckType.OT2_STANDARD,
        requested_model_by_id={},
        load_location_by_module_id={},
        hardware_by_module_id={},
        substate_by_module_id={},
        module_offset_by_serial={},
        additional_slots_occupied_by_module_id={},
        deck_fixed_labware=[],
    )


@pytest.mark.parametrize(
    argnames=["module_definition", "params_model", "result_model", "expected_substate"],
    argvalues=[
        (
            lazy_fixture("magdeck_v2_def"),
            ModuleModel.MAGNETIC_MODULE_V2,
            ModuleModel.MAGNETIC_MODULE_V2,
            MagneticModuleSubState(
                module_id=MagneticModuleId("module-id"),
                model=ModuleModel.MAGNETIC_MODULE_V2,
            ),
        ),
        (
            lazy_fixture("heater_shaker_v1_def"),
            ModuleModel.HEATER_SHAKER_MODULE_V1,
            ModuleModel.HEATER_SHAKER_MODULE_V1,
            HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId("module-id"),
                labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
                is_plate_shaking=False,
                plate_target_temperature=None,
            ),
        ),
        (
            lazy_fixture("tempdeck_v1_def"),
            ModuleModel.TEMPERATURE_MODULE_V1,
            ModuleModel.TEMPERATURE_MODULE_V1,
            TemperatureModuleSubState(
                module_id=TemperatureModuleId("module-id"),
                plate_target_temperature=None,
            ),
        ),
        (
            lazy_fixture("tempdeck_v1_def"),
            ModuleModel.TEMPERATURE_MODULE_V2,
            ModuleModel.TEMPERATURE_MODULE_V1,
            TemperatureModuleSubState(
                module_id=TemperatureModuleId("module-id"),
                plate_target_temperature=None,
            ),
        ),
        (
            lazy_fixture("tempdeck_v2_def"),
            ModuleModel.TEMPERATURE_MODULE_V1,
            ModuleModel.TEMPERATURE_MODULE_V2,
            TemperatureModuleSubState(
                module_id=TemperatureModuleId("module-id"),
                plate_target_temperature=None,
            ),
        ),
        (
            lazy_fixture("tempdeck_v2_def"),
            ModuleModel.TEMPERATURE_MODULE_V2,
            ModuleModel.TEMPERATURE_MODULE_V2,
            TemperatureModuleSubState(
                module_id=TemperatureModuleId("module-id"),
                plate_target_temperature=None,
            ),
        ),
        (
            lazy_fixture("thermocycler_v1_def"),
            ModuleModel.THERMOCYCLER_MODULE_V1,
            ModuleModel.THERMOCYCLER_MODULE_V1,
            ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId("module-id"),
                is_lid_open=False,
                target_block_temperature=None,
                target_lid_temperature=None,
            ),
        ),
    ],
)
def test_load_module(
    module_definition: ModuleDefinition,
    params_model: ModuleModel,
    result_model: ModuleModel,
    expected_substate: ModuleSubStateType,
) -> None:
    """It should handle a successful LoadModule command."""
    action = actions.SucceedCommandAction(
        command=commands.LoadModule.model_construct(  # type: ignore[call-arg]
            params=commands.LoadModuleParams(
                model=params_model,
                location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
            ),
            result=commands.LoadModuleResult(
                moduleId="module-id",
                model=result_model,
                serialNumber="serial-number",
            ),
        ),
        state_update=update_types.StateUpdate().set_load_module(
            module_id="module-id",
            definition=module_definition,
            requested_model=params_model,
            serial_number="serial-number",
            slot_name=DeckSlotName.SLOT_1,
        ),
    )

    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )
    subject.handle_action(action)

    assert subject.state == ModuleState(
        deck_type=DeckType.OT2_STANDARD,
        load_location_by_module_id={
            "module-id": deck_configuration_provider.get_cutout_id_by_deck_slot_name(
                DeckSlotName.SLOT_1
            )
        },
        requested_model_by_id={"module-id": params_model},
        hardware_by_module_id={
            "module-id": HardwareModule(
                serial_number="serial-number",
                definition=module_definition,
            )
        },
        substate_by_module_id={"module-id": expected_substate},
        module_offset_by_serial={},
        additional_slots_occupied_by_module_id={},
        deck_fixed_labware=[],
    )


@pytest.mark.parametrize(
    argnames=["tc_slot", "deck_type", "robot_type", "expected_additional_slots"],
    argvalues=[
        (
            DeckSlotName.SLOT_7,
            DeckType.OT2_STANDARD,
            "OT-2 Standard",
            [DeckSlotName.SLOT_8, DeckSlotName.SLOT_10, DeckSlotName.SLOT_11],
        ),
        (
            DeckSlotName.SLOT_B1,
            DeckType.OT3_STANDARD,
            "OT-3 Standard",
            [DeckSlotName.SLOT_A1],
        ),
    ],
)
def test_load_thermocycler_in_thermocycler_slot(
    tc_slot: DeckSlotName,
    deck_type: DeckType,
    robot_type: RobotType,
    expected_additional_slots: List[DeckSlotName],
    thermocycler_v2_def: ModuleDefinition,
) -> None:
    """It should update additional slots for thermocycler module."""
    action = actions.SucceedCommandAction(
        command=commands.LoadModule.model_construct(  # type: ignore[call-arg]
            params=commands.LoadModuleParams(
                model=ModuleModel.THERMOCYCLER_MODULE_V2,
                location=DeckSlotLocation(slotName=tc_slot),
            ),
            result=commands.LoadModuleResult(
                moduleId="module-id",
                model=ModuleModel.THERMOCYCLER_MODULE_V2,
                serialNumber="serial-number",
            ),
        ),
        state_update=update_types.StateUpdate().set_load_module(
            module_id="module-id",
            definition=thermocycler_v2_def,
            requested_model=ModuleModel.THERMOCYCLER_MODULE_V2,
            serial_number="serial-number",
            slot_name=tc_slot,
        ),
    )
    load_position_for_module = f"cutout{tc_slot.value}"

    subject = ModuleStore(
        Config(
            use_simulated_deck_config=False,
            robot_type=robot_type,
            deck_type=deck_type,
        ),
        deck_fixed_labware=[],
    )
    subject.handle_action(action)

    assert subject.state.load_location_by_module_id == {
        "module-id": load_position_for_module
    }
    assert subject.state.additional_slots_occupied_by_module_id == {
        "module-id": expected_additional_slots
    }


@pytest.mark.parametrize(
    argnames=["module_definition", "live_data", "expected_substate"],
    argvalues=[
        (
            lazy_fixture("magdeck_v2_def"),
            {},
            MagneticModuleSubState(
                module_id=MagneticModuleId("module-id"),
                model=ModuleModel.MAGNETIC_MODULE_V2,
            ),
        ),
        (
            lazy_fixture("heater_shaker_v1_def"),
            {
                "status": "abc",
                "data": {
                    "labwareLatchStatus": "idle_closed",
                    "speedStatus": "holding at target",
                    "targetSpeed": 123,
                    "targetTemp": 123,
                },
            },
            HeaterShakerModuleSubState(
                module_id=HeaterShakerModuleId("module-id"),
                labware_latch_status=HeaterShakerLatchStatus.CLOSED,
                is_plate_shaking=True,
                plate_target_temperature=123,
            ),
        ),
        (
            lazy_fixture("tempdeck_v2_def"),
            {"status": "abc", "data": {"targetTemp": 123}},
            TemperatureModuleSubState(
                module_id=TemperatureModuleId("module-id"),
                plate_target_temperature=123,
            ),
        ),
        (
            lazy_fixture("thermocycler_v1_def"),
            {
                "status": "abc",
                "data": {
                    "targetTemp": 123,
                    "lidTarget": 321,
                    "lid": "open",
                },
            },
            ThermocyclerModuleSubState(
                module_id=ThermocyclerModuleId("module-id"),
                is_lid_open=True,
                target_block_temperature=123,
                target_lid_temperature=321,
            ),
        ),
    ],
)
def test_add_module_action(
    module_definition: ModuleDefinition,
    live_data: LiveData,
    expected_substate: ModuleSubStateType,
) -> None:
    """It should be able to add attached modules directly into state."""
    action = actions.AddModuleAction(
        module_id="module-id",
        serial_number="serial-number",
        definition=module_definition,
        module_live_data=live_data,
    )

    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )
    subject.handle_action(action)

    assert subject.state == ModuleState(
        deck_type=DeckType.OT2_STANDARD,
        load_location_by_module_id={"module-id": None},
        requested_model_by_id={"module-id": None},
        hardware_by_module_id={
            "module-id": HardwareModule(
                serial_number="serial-number",
                definition=module_definition,
            )
        },
        substate_by_module_id={"module-id": expected_substate},
        module_offset_by_serial={},
        additional_slots_occupied_by_module_id={},
        deck_fixed_labware=[],
    )


def test_handle_hs_temperature_commands(heater_shaker_v1_def: ModuleDefinition) -> None:
    """It should update `plate_target_temperature` correctly."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            serialNumber="serial-number",
        ),
    )
    set_temp_cmd = hs_commands.SetTargetTemperature.model_construct(  # type: ignore[call-arg]
        params=hs_commands.SetTargetTemperatureParams(moduleId="module-id", celsius=42),
        result=hs_commands.SetTargetTemperatureResult(taskId="taskId"),
    )
    deactivate_cmd = hs_commands.DeactivateHeater.model_construct(  # type: ignore[call-arg]
        params=hs_commands.DeactivateHeaterParams(moduleId="module-id"),
        result=hs_commands.DeactivateHeaterResult(),
    )
    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=heater_shaker_v1_def,
                requested_model=ModuleModel.HEATER_SHAKER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(actions.SucceedCommandAction(command=set_temp_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=False,
            plate_target_temperature=42,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=deactivate_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    }


def test_handle_hs_shake_commands(heater_shaker_v1_def: ModuleDefinition) -> None:
    """It should update heater-shaker's `is_plate_shaking` correctly."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            serialNumber="serial-number",
        ),
    )
    set_shake_cmd = hs_commands.SetAndWaitForShakeSpeed.model_construct(  # type: ignore[call-arg]
        params=hs_commands.SetAndWaitForShakeSpeedParams(moduleId="module-id", rpm=111),
        result=hs_commands.SetAndWaitForShakeSpeedResult(pipetteRetracted=False),
    )
    start_set_shake_cmd = hs_commands.SetShakeSpeed.model_construct(  # type: ignore[call-arg]
        params=hs_commands.SetShakeSpeedParams(moduleId="module-id", rpm=111),
        result=hs_commands.SetShakeSpeedResult(pipetteRetracted=False, taskId="taskId"),
    )
    deactivate_cmd = hs_commands.DeactivateShaker.model_construct(  # type: ignore[call-arg]
        params=hs_commands.DeactivateShakerParams(moduleId="module-id"),
        result=hs_commands.DeactivateShakerResult(),
    )

    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=heater_shaker_v1_def,
                requested_model=ModuleModel.HEATER_SHAKER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(actions.SucceedCommandAction(command=set_shake_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=True,
            plate_target_temperature=None,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=start_set_shake_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=True,
            plate_target_temperature=None,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=deactivate_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    }


def test_handle_hs_labware_latch_commands(
    heater_shaker_v1_def: ModuleDefinition,
) -> None:
    """It should update heater-shaker's `is_labware_latch_closed` correctly."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.HEATER_SHAKER_MODULE_V1,
            serialNumber="serial-number",
        ),
    )
    close_latch_cmd = hs_commands.CloseLabwareLatch.model_construct(  # type: ignore[call-arg]
        params=hs_commands.CloseLabwareLatchParams(moduleId="module-id"),
        result=hs_commands.CloseLabwareLatchResult(),
    )
    open_latch_cmd = hs_commands.OpenLabwareLatch.model_construct(  # type: ignore[call-arg]
        params=hs_commands.OpenLabwareLatchParams(moduleId="module-id"),
        result=hs_commands.OpenLabwareLatchResult(pipetteRetracted=False),
    )
    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=heater_shaker_v1_def,
                requested_model=ModuleModel.HEATER_SHAKER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.UNKNOWN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    }

    subject.handle_action(actions.SucceedCommandAction(command=close_latch_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.CLOSED,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=open_latch_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": HeaterShakerModuleSubState(
            module_id=HeaterShakerModuleId("module-id"),
            labware_latch_status=HeaterShakerLatchStatus.OPEN,
            is_plate_shaking=False,
            plate_target_temperature=None,
        )
    }


def test_handle_tempdeck_temperature_commands(
    tempdeck_v2_def: ModuleDefinition,
) -> None:
    """It should update Tempdeck's `plate_target_temperature` correctly."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.TEMPERATURE_MODULE_V2,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.TEMPERATURE_MODULE_V2,
            serialNumber="serial-number",
        ),
    )
    set_temp_cmd = temp_commands.SetTargetTemperature.model_construct(  # type: ignore[call-arg]
        params=temp_commands.SetTargetTemperatureParams(
            moduleId="module-id", celsius=42.4
        ),
        result=temp_commands.SetTargetTemperatureResult(
            targetTemperature=42, taskId="taskId"
        ),
    )
    deactivate_cmd = temp_commands.DeactivateTemperature.model_construct(  # type: ignore[call-arg]
        params=temp_commands.DeactivateTemperatureParams(moduleId="module-id"),
        result=temp_commands.DeactivateTemperatureResult(),
    )
    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=tempdeck_v2_def,
                requested_model=ModuleModel.HEATER_SHAKER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(actions.SucceedCommandAction(command=set_temp_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": TemperatureModuleSubState(
            module_id=TemperatureModuleId("module-id"), plate_target_temperature=42
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=deactivate_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": TemperatureModuleSubState(
            module_id=TemperatureModuleId("module-id"), plate_target_temperature=None
        )
    }


def test_handle_thermocycler_temperature_commands(
    thermocycler_v1_def: ModuleDefinition,
) -> None:
    """It should update thermocycler's temperature statuses correctly."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.THERMOCYCLER_MODULE_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.THERMOCYCLER_MODULE_V1,
            serialNumber="serial-number",
        ),
    )
    set_block_temp_cmd = tc_commands.SetTargetBlockTemperature.model_construct(  # type: ignore[call-arg]
        params=tc_commands.SetTargetBlockTemperatureParams(
            moduleId="module-id", celsius=42.4
        ),
        result=tc_commands.SetTargetBlockTemperatureResult(targetBlockTemperature=42.4),
    )
    deactivate_block_cmd = tc_commands.DeactivateBlock.model_construct(  # type: ignore[call-arg]
        params=tc_commands.DeactivateBlockParams(moduleId="module-id"),
        result=tc_commands.DeactivateBlockResult(),
    )
    set_lid_temp_cmd = tc_commands.SetTargetLidTemperature.model_construct(  # type: ignore[call-arg]
        params=tc_commands.SetTargetLidTemperatureParams(
            moduleId="module-id", celsius=35.3
        ),
        result=tc_commands.SetTargetLidTemperatureResult(targetLidTemperature=35.3),
    )
    deactivate_lid_cmd = tc_commands.DeactivateLid.model_construct(  # type: ignore[call-arg]
        params=tc_commands.DeactivateLidParams(moduleId="module-id"),
        result=tc_commands.DeactivateLidResult(),
    )
    subject = ModuleStore(
        config=_OT2_STANDARD_CONFIG,
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=thermocycler_v1_def,
                requested_model=ModuleModel.THERMOCYCLER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(actions.SucceedCommandAction(command=set_block_temp_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_block_temperature=42.4,
            target_lid_temperature=None,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=set_lid_temp_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_block_temperature=42.4,
            target_lid_temperature=35.3,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=deactivate_lid_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_block_temperature=42.4,
            target_lid_temperature=None,
        )
    }
    subject.handle_action(actions.SucceedCommandAction(command=deactivate_block_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_block_temperature=None,
            target_lid_temperature=None,
        )
    }


def test_handle_thermocycler_lid_commands(
    thermocycler_v1_def: ModuleDefinition,
) -> None:
    """It should update thermocycler's lid status after executing lid commands."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.THERMOCYCLER_MODULE_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.THERMOCYCLER_MODULE_V1,
            serialNumber="serial-number",
        ),
    )

    open_lid_cmd = tc_commands.OpenLid.model_construct(  # type: ignore[call-arg]
        params=tc_commands.OpenLidParams(moduleId="module-id"),
        result=tc_commands.OpenLidResult(),
    )
    close_lid_cmd = tc_commands.CloseLid.model_construct(  # type: ignore[call-arg]
        params=tc_commands.CloseLidParams(moduleId="module-id"),
        result=tc_commands.CloseLidResult(),
    )

    subject = ModuleStore(
        Config(
            use_simulated_deck_config=False,
            robot_type="OT-3 Standard",
            deck_type=DeckType.OT3_STANDARD,
        ),
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=thermocycler_v1_def,
                requested_model=ModuleModel.THERMOCYCLER_MODULE_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(actions.SucceedCommandAction(command=open_lid_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=True,
            target_block_temperature=None,
            target_lid_temperature=None,
        )
    }

    subject.handle_action(actions.SucceedCommandAction(command=close_lid_cmd))
    assert subject.state.substate_by_module_id == {
        "module-id": ThermocyclerModuleSubState(
            module_id=ThermocyclerModuleId("module-id"),
            is_lid_open=False,
            target_block_temperature=None,
            target_lid_temperature=None,
        )
    }


def test_handle_absorbance_reader_commands(
    abs_reader_v1_def: ModuleDefinition,
) -> None:
    """It should update absorbance reader state."""
    load_module_cmd = commands.LoadModule.model_construct(  # type: ignore[call-arg]
        params=commands.LoadModuleParams(
            model=ModuleModel.ABSORBANCE_READER_V1,
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_1),
        ),
        result=commands.LoadModuleResult(
            moduleId="module-id",
            model=ModuleModel.ABSORBANCE_READER_V1,
            serialNumber="serial-number",
        ),
    )

    initialize_reader = commands.Comment.model_construct(  # type: ignore[call-arg]
        params=commands.CommentParams(message="hello"),
        result=commands.CommentResult(),
    )
    open_lid = commands.Comment.model_construct(  # type: ignore[call-arg]
        params=commands.CommentParams(message="hello dude"),
        result=commands.CommentResult(),
    )

    read_data = commands.Comment.model_construct(  # type: ignore[call-arg]
        params=commands.CommentParams(message="hello man"),
        result=commands.CommentResult(),
    )

    close_lid = commands.Comment.model_construct(  # type: ignore[call-arg]
        params=commands.CommentParams(message="hello ladies"),
        result=commands.CommentResult(),
    )

    subject = ModuleStore(
        Config(
            use_simulated_deck_config=False,
            robot_type="OT-3 Standard",
            deck_type=DeckType.OT3_STANDARD,
        ),
        deck_fixed_labware=[],
    )

    subject.handle_action(
        actions.SucceedCommandAction(
            command=load_module_cmd,
            state_update=update_types.StateUpdate().set_load_module(
                module_id="module-id",
                definition=abs_reader_v1_def,
                requested_model=ModuleModel.ABSORBANCE_READER_V1,
                serial_number="serial-number",
                slot_name=DeckSlotName.SLOT_1,
            ),
        )
    )
    subject.handle_action(
        actions.SucceedCommandAction(
            command=initialize_reader,
            state_update=update_types.StateUpdate().initialize_absorbance_reader(
                "module-id", "single", [1], None
            ),
        )
    )
    assert subject.state.substate_by_module_id == {
        "module-id": AbsorbanceReaderSubState(
            module_id=AbsorbanceReaderId("module-id"),
            is_lid_on=True,
            configured=True,
            measured=False,
            data=None,
            configured_wavelengths=[1],
            measure_mode="single",  # type: ignore[arg-type]
            reference_wavelength=None,
        )
    }

    subject.handle_action(
        actions.SucceedCommandAction(
            command=open_lid,
            state_update=update_types.StateUpdate().set_absorbance_reader_lid(
                module_id="module-id", is_lid_on=False
            ),
        )
    )
    assert subject.state.substate_by_module_id == {
        "module-id": AbsorbanceReaderSubState(
            module_id=AbsorbanceReaderId("module-id"),
            is_lid_on=False,
            configured=True,
            measured=True,
            data=None,
            configured_wavelengths=[1],
            measure_mode="single",  # type: ignore[arg-type]
            reference_wavelength=None,
        )
    }

    subject.handle_action(
        actions.SucceedCommandAction(
            command=read_data,
            state_update=update_types.StateUpdate().set_absorbance_reader_data(
                module_id="module-id", read_result={1: {"A1": 1.2}}
            ),
        )
    )
    assert subject.state.substate_by_module_id == {
        "module-id": AbsorbanceReaderSubState(
            module_id=AbsorbanceReaderId("module-id"),
            is_lid_on=False,
            configured=True,
            measured=True,
            data={1: {"A1": 1.2}},
            configured_wavelengths=[1],
            measure_mode="single",  # type: ignore[arg-type]
            reference_wavelength=None,
        )
    }

    subject.handle_action(
        actions.SucceedCommandAction(
            command=close_lid,
            state_update=update_types.StateUpdate().set_absorbance_reader_lid(
                module_id="module-id", is_lid_on=True
            ),
        )
    )
    assert subject.state.substate_by_module_id == {
        "module-id": AbsorbanceReaderSubState(
            module_id=AbsorbanceReaderId("module-id"),
            is_lid_on=True,
            configured=True,
            measured=True,
            data={1: {"A1": 1.2}},
            configured_wavelengths=[1],
            measure_mode="single",  # type: ignore[arg-type]
            reference_wavelength=None,
        )
    }



"""State test fixtures."""

import pytest
from decoy import Decoy

from opentrons.protocol_engine.state.addressable_areas import AddressableAreaView
from opentrons.protocol_engine.state.geometry import GeometryView
from opentrons.protocol_engine.state.labware import LabwareView
from opentrons.protocol_engine.state.pipettes import PipetteView


@pytest.fixture
def labware_view(decoy: Decoy) -> LabwareView:
    """Get a mock in the shape of a LabwareView."""
    return decoy.mock(cls=LabwareView)


@pytest.fixture
def pipette_view(decoy: Decoy) -> PipetteView:
    """Get a mock in the shape of a PipetteView."""
    return decoy.mock(cls=PipetteView)


@pytest.fixture
def addressable_area_view(decoy: Decoy) -> AddressableAreaView:
    """Get a mock in the shape of a AddressableAreaView."""
    return decoy.mock(cls=AddressableAreaView)


@pytest.fixture
def geometry_view(decoy: Decoy) -> GeometryView:
    """Get a mock in the shape of a GeometryView."""
    return decoy.mock(cls=GeometryView)



"""Tests for Protocol API input validation."""

import re
from contextlib import nullcontext as do_not_raise
from typing import Any, ContextManager, Dict, List, Optional, Sequence, Type, Union

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import (
    LabwareDefinition2,
    LabwareRole,
)
from opentrons_shared_data.labware.labware_definition import (
    Parameters2 as LabwareDefinition2Parameters,
)
from opentrons_shared_data.pipette.types import PipetteNameType
from opentrons_shared_data.robot.types import RobotType

from opentrons.hardware_control.modules.types import (
    HeaterShakerModuleModel,
    MagneticModuleModel,
    ModuleModel,
    TemperatureModuleModel,
    ThermocyclerModuleModel,
    ThermocyclerStep,
)
from opentrons.protocol_api import (
    Labware,
    TrashBin,
    WasteChute,
    Well,
)
from opentrons.protocol_api import (
    validation as subject,
)
from opentrons.protocols.advanced_control.transfers.common import TransferTipPolicyV2
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import (
    AxisMapType,
    AxisType,
    DeckSlotName,
    Location,
    Mount,
    Point,
    StagingSlotName,
    StringAxisMap,
)


@pytest.mark.parametrize(
    ["input_mount", "input_pipette", "expected"],
    [
        # Different string capitalizations:
        ("left", PipetteNameType.P300_MULTI_GEN2, Mount.LEFT),
        ("right", PipetteNameType.P300_MULTI_GEN2, Mount.RIGHT),
        ("LeFt", PipetteNameType.P300_MULTI_GEN2, Mount.LEFT),
        # Passing in a Mount:
        (Mount.LEFT, PipetteNameType.P300_MULTI_GEN2, Mount.LEFT),
        (Mount.RIGHT, PipetteNameType.P300_MULTI_GEN2, Mount.RIGHT),
        # Special handling for the 96-channel:
        ("left", PipetteNameType.P1000_96, Mount.LEFT),
        ("right", PipetteNameType.P1000_96, Mount.LEFT),
        (None, PipetteNameType.P1000_96, Mount.LEFT),
    ],
)
def test_ensure_mount(
    input_mount: Union[str, Mount, None],
    input_pipette: PipetteNameType,
    expected: Mount,
) -> None:
    """It should properly map strings and mounts."""
    result = subject.ensure_mount_for_pipette(input_mount, input_pipette)
    assert result == expected


def test_ensure_mount_input_invalid() -> None:
    """It should raise if given invalid mount input."""
    with pytest.raises(
        subject.InvalidPipetteMountError, match="must be 'left' or 'right'"
    ):
        subject.ensure_mount_for_pipette("oh no", PipetteNameType.P300_MULTI_GEN2)

    # Any mount is valid for the 96-Channel, but it needs to be a valid mount.
    with pytest.raises(
        subject.InvalidPipetteMountError, match="must be 'left' or 'right'"
    ):
        subject.ensure_mount_for_pipette("oh no", PipetteNameType.P1000_96)

    with pytest.raises(
        subject.PipetteMountTypeError,
        match="'left', 'right', or an opentrons.types.Mount",
    ):
        subject.ensure_mount_for_pipette(42, PipetteNameType.P300_MULTI_GEN2)  # type: ignore[arg-type]

    with pytest.raises(
        subject.InvalidPipetteMountError, match="Use the left or right mounts instead"
    ):
        subject.ensure_mount_for_pipette(
            Mount.EXTENSION, PipetteNameType.P300_MULTI_GEN2
        )

    with pytest.raises(
        subject.InvalidPipetteMountError, match="You must specify a left or right mount"
    ):
        subject.ensure_mount_for_pipette(None, PipetteNameType.P300_MULTI_GEN2)


@pytest.mark.parametrize(
    ["input_value", "expected"],
    [
        # Every OT-2 pipette:
        ("p10_single", PipetteNameType.P10_SINGLE),
        ("p10_multi", PipetteNameType.P10_MULTI),
        ("p50_single", PipetteNameType.P50_SINGLE),
        ("p50_multi", PipetteNameType.P50_MULTI),
        ("p300_single", PipetteNameType.P300_SINGLE),
        ("p300_multi", PipetteNameType.P300_MULTI),
        ("p1000_single", PipetteNameType.P1000_SINGLE),
        ("p20_single_gen2", PipetteNameType.P20_SINGLE_GEN2),
        ("p20_multi_gen2", PipetteNameType.P20_MULTI_GEN2),
        ("p300_single_gen2", PipetteNameType.P300_SINGLE_GEN2),
        ("p300_multi_gen2", PipetteNameType.P300_MULTI_GEN2),
        ("p1000_single_gen2", PipetteNameType.P1000_SINGLE_GEN2),
        # Every Flex pipette:
        ("flex_1channel_50", PipetteNameType.P50_SINGLE_FLEX),
        ("flex_8channel_50", PipetteNameType.P50_MULTI_FLEX),
        ("flex_1channel_1000", PipetteNameType.P1000_SINGLE_FLEX),
        ("flex_8channel_1000", PipetteNameType.P1000_MULTI_FLEX),
        ("flex_96channel_1000", PipetteNameType.P1000_96),
        # Weird capitalization:
        ("P300_muLTI_gen2", PipetteNameType.P300_MULTI_GEN2),
    ],
)
def test_ensure_pipette_name(input_value: str, expected: PipetteNameType) -> None:
    """It should properly map strings and PipetteNameType enums."""
    result = subject.ensure_pipette_name(input_value)
    assert result == expected


@pytest.mark.parametrize(
    "input_value",
    [
        "oh-no",  # Not even remotely a pipette name.
        "p1000_single_gen3",  # Obsolete name for Flex pipette.
        "p1000_single_flex",  # Internal-only name for Flex pipette.
        "p1000_96",  # Internal-only name for Flex pipette.
    ],
)
def test_ensure_pipette_input_invalid(input_value: str) -> None:
    """It should raise a ValueError if given an invalid name."""
    with pytest.raises(
        ValueError,
        match=f"Cannot resolve {input_value} to pipette, must be given valid pipette name",
    ):
        subject.ensure_pipette_name(input_value)


@pytest.mark.parametrize(
    ["input_value", "input_api_version", "input_robot_type", "expected"],
    [
        # Integer or integer-as-string slots:
        ("1", APIVersion(2, 0), "OT-2 Standard", DeckSlotName.SLOT_1),
        ("1", APIVersion(2, 0), "OT-3 Standard", DeckSlotName.SLOT_D1),
        (1, APIVersion(2, 0), "OT-2 Standard", DeckSlotName.SLOT_1),
        (1, APIVersion(2, 0), "OT-3 Standard", DeckSlotName.SLOT_D1),
        ("12", APIVersion(2, 0), "OT-2 Standard", DeckSlotName.FIXED_TRASH),
        (12, APIVersion(2, 0), "OT-3 Standard", DeckSlotName.SLOT_A3),
        # Coordinate slots:
        ("d1", APIVersion(2, 15), "OT-2 Standard", DeckSlotName.SLOT_1),
        ("d1", APIVersion(2, 15), "OT-3 Standard", DeckSlotName.SLOT_D1),
        ("D1", APIVersion(2, 15), "OT-2 Standard", DeckSlotName.SLOT_1),
        ("D1", APIVersion(2, 15), "OT-3 Standard", DeckSlotName.SLOT_D1),
        ("a3", APIVersion(2, 15), "OT-2 Standard", DeckSlotName.FIXED_TRASH),
        ("a3", APIVersion(2, 15), "OT-3 Standard", DeckSlotName.SLOT_A3),
        ("A3", APIVersion(2, 15), "OT-2 Standard", DeckSlotName.FIXED_TRASH),
        ("A3", APIVersion(2, 15), "OT-3 Standard", DeckSlotName.SLOT_A3),
        # Staging slots:
        ("A4", APIVersion(2, 16), "OT-3 Standard", StagingSlotName.SLOT_A4),
        ("b4", APIVersion(2, 16), "OT-3 Standard", StagingSlotName.SLOT_B4),
        ("C4", APIVersion(2, 16), "OT-3 Standard", StagingSlotName.SLOT_C4),
        ("d4", APIVersion(2, 16), "OT-3 Standard", StagingSlotName.SLOT_D4),
    ],
)
def test_ensure_and_convert_deck_slot(
    input_value: Union[str, int],
    input_api_version: APIVersion,
    input_robot_type: RobotType,
    expected: DeckSlotName,
) -> None:
    """It should map strings and ints to DeckSlotName values."""
    result = subject.ensure_and_convert_deck_slot(
        input_value, input_api_version, input_robot_type
    )
    assert result == expected


@pytest.mark.parametrize(
    ["input_value", "input_api_version", "expected_error_type", "expected_error_match"],
    [
        ("0", APIVersion(2, 0), ValueError, "not a valid deck slot"),
        (0, APIVersion(2, 0), ValueError, "not a valid deck slot"),
        ("13", APIVersion(2, 0), ValueError, "not a valid deck slot"),
        (13, APIVersion(2, 0), ValueError, "not a valid deck slot"),
        ("b7", APIVersion(2, 0), ValueError, "not a valid deck slot"),
        ("B7", APIVersion(2, 0), ValueError, "not a valid deck slot"),
        (1.23, APIVersion(2, 0), TypeError, "must be a string or integer"),
        (
            "A1",
            APIVersion(2, 0),
            APIVersionError,
            re.escape(
                "Error 4011 INCORRECT_API_VERSION (APIVersionError): Specifying a deck slot like 'A1' is not available until API version 2.15. You are currently using API version 2.0. Increase your protocol's apiLevel, or use slot '10' instead."
            ),
        ),
        ("A4", APIVersion(2, 15), APIVersionError, "Using a staging deck slot"),
    ],
)
@pytest.mark.parametrize("input_robot_type", ["OT-2 Standard", "OT-3 Standard"])
def test_ensure_deck_slot_invalid(
    input_value: object,
    input_api_version: APIVersion,
    input_robot_type: RobotType,
    expected_error_type: Type[Exception],
    expected_error_match: str,
) -> None:
    """It should raise an exception if given an invalid name."""
    with pytest.raises(expected_error_type, match=expected_error_match):
        subject.ensure_and_convert_deck_slot(
            input_value,  # type: ignore[arg-type]
            input_api_version,
            input_robot_type,
        )


def test_ensure_lowercase_name() -> None:
    """It should properly map strings and PipetteNameType enums."""
    result = subject.ensure_lowercase_name("aBcDeFg")
    assert result == "abcdefg"


def test_ensure_lowercase_name_invalid() -> None:
    """It should raise a ValueError if given an invalid name."""
    with pytest.raises(TypeError, match="must be a string"):
        subject.ensure_lowercase_name(101)  # type: ignore[arg-type]


@pytest.mark.parametrize(
    ("definition", "expected_raise"),
    [
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            do_not_raise(),
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            do_not_raise(),
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            pytest.raises(subject.LabwareDefinitionIsNotLabwareError),
        ),
    ],
)
def test_ensure_definition_is_labware(
    definition: LabwareDefinition2, expected_raise: ContextManager[Any]
) -> None:
    """It should check if the Labware Definition is defined as a regular labware."""
    with expected_raise:
        subject.ensure_definition_is_labware(definition)


@pytest.mark.parametrize(
    ("definition", "expected_raise"),
    [
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.adapter],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            do_not_raise(),
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            pytest.raises(subject.LabwareDefinitionIsNotAdapterError),
        ),
        (
            LabwareDefinition2.model_construct(  # type: ignore[call-arg]
                allowedRoles=[LabwareRole.labware],
                parameters=LabwareDefinition2Parameters.model_construct(loadName="Foo"),  # type: ignore[call-arg]
            ),
            pytest.raises(subject.LabwareDefinitionIsNotAdapterError),
        ),
    ],
)
def test_ensure_definition_is_adapter(
    definition: LabwareDefinition2, expected_raise: ContextManager[Any]
) -> None:
    """It should check if the Labware Definition is defined as an adapter."""
    with expected_raise:
        subject.ensure_definition_is_adapter(definition)


@pytest.mark.parametrize(
    ("load_name", "expected_model"),
    [
        ("magdeck", MagneticModuleModel.MAGNETIC_V1),
        ("MaGdEcK", MagneticModuleModel.MAGNETIC_V1),
        ("magnetic module", MagneticModuleModel.MAGNETIC_V1),
        ("magneticModuleV1", MagneticModuleModel.MAGNETIC_V1),
        ("magnetic module gen2", MagneticModuleModel.MAGNETIC_V2),
        ("magneticModuleV2", MagneticModuleModel.MAGNETIC_V2),
        ("tempdeck", TemperatureModuleModel.TEMPERATURE_V1),
        ("tEmpDeCk", TemperatureModuleModel.TEMPERATURE_V1),
        ("temperatureModuleV1", TemperatureModuleModel.TEMPERATURE_V1),
        ("temperature module", TemperatureModuleModel.TEMPERATURE_V1),
        ("temperature module gen2", TemperatureModuleModel.TEMPERATURE_V2),
        ("temperatureModuleV2", TemperatureModuleModel.TEMPERATURE_V2),
        ("thermocycler", ThermocyclerModuleModel.THERMOCYCLER_V1),
        ("ThErMoCyClEr", ThermocyclerModuleModel.THERMOCYCLER_V1),
        ("thermocycler module", ThermocyclerModuleModel.THERMOCYCLER_V1),
        ("thermocyclerModuleV1", ThermocyclerModuleModel.THERMOCYCLER_V1),
        ("thermocycler module gen2", ThermocyclerModuleModel.THERMOCYCLER_V2),
        ("thermocyclerModuleV2", ThermocyclerModuleModel.THERMOCYCLER_V2),
        ("heaterShakerModuleV1", HeaterShakerModuleModel.HEATER_SHAKER_V1),
    ],
)
def test_ensure_module_model(load_name: str, expected_model: ModuleModel) -> None:
    """It should map an module load name to a specific model."""
    result = subject.ensure_module_model(load_name)
    assert result == expected_model


def test_ensure_module_model_invalid() -> None:
    """It should reject invalid module load names."""
    with pytest.raises(ValueError, match="not a valid module load name"):
        subject.ensure_module_model("spline reticulator")

    with pytest.raises(TypeError, match="must be a string"):
        subject.ensure_module_model(42)  # type: ignore[arg-type]


@pytest.mark.parametrize(
    ["seconds", "minutes", "expected"],
    [
        (42.42, None, 42.42),
        (None, 1.2, 72.0),
        (42.42, 1.2, 114.42),
        (None, None, 0),
    ],
)
def test_ensure_hold_time_seconds(
    seconds: Optional[float], minutes: Optional[float], expected: float
) -> None:
    """It should ensure hold time is in seconds only."""
    result = subject.ensure_hold_time_seconds(seconds=seconds, minutes=minutes)
    assert result == expected


@pytest.mark.parametrize(
    ["repetitions", "expected"],
    [
        (1, 1),
        (2, 2),
        (999, 999),
    ],
)
def test_ensure_thermocycler_repetition_count(repetitions: int, expected: int) -> None:
    """It should return a given positive integer."""
    result = subject.ensure_thermocycler_repetition_count(repetitions)
    assert result == expected


@pytest.mark.parametrize(
    "repetitions",
    [
        0,
        -1,
        -999,
    ],
)
def test_ensure_thermocycler_repetition_count_raises(repetitions: int) -> None:
    """It should raise if repetitions is zero or negative."""
    with pytest.raises(ValueError):
        subject.ensure_thermocycler_repetition_count(repetitions)


@pytest.mark.parametrize(
    ["steps", "expected"],
    [
        (
            [
                {
                    "temperature": 42.0,
                    "hold_time_minutes": 12.3,
                    "hold_time_seconds": 45.6,
                }
            ],
            [{"temperature": 42.0, "hold_time_seconds": 783.6, "ramp_rate": None}],
        ),
        (
            [{"temperature": 42.0, "hold_time_seconds": 45.6, "ramp_rate": 1.0}],
            [{"temperature": 42.0, "hold_time_seconds": 45.6, "ramp_rate": 1.0}],
        ),
        (
            [{"temperature": 42.0, "hold_time_minutes": 12.3}],
            [{"temperature": 42.0, "hold_time_seconds": 738.0, "ramp_rate": None}],
        ),
        (
            [
                {"temperature": 42.0, "hold_time_seconds": 12.3},
                {"temperature": 52.0, "hold_time_minutes": 12.3},
            ],
            [
                {"temperature": 42.0, "hold_time_seconds": 12.3, "ramp_rate": None},
                {"temperature": 52.0, "hold_time_seconds": 738.0, "ramp_rate": None},
            ],
        ),
        ([], []),
    ],
)
def test_ensure_thermocycler_profile_steps(
    steps: List[ThermocyclerStep], expected: List[ThermocyclerStep]
) -> None:
    """It should ensure thermocycler profile steps are valid and hold time is expressed in seconds only."""
    result = subject.ensure_thermocycler_profile_steps(steps)
    assert result == expected


@pytest.mark.parametrize(
    "steps",
    [
        [{"hold_time_minutes": 12.3, "hold_time_seconds": 45.6}],
        [{"temperature": 42.0}],
    ],
)
def test_ensure_thermocycler_profile_steps_invalid(
    steps: List[ThermocyclerStep],
) -> None:
    """It should raise a ValueError when given invalid thermocycler profile steps."""
    with pytest.raises(ValueError):
        subject.ensure_thermocycler_profile_steps(steps)


@pytest.mark.parametrize("offset", [{}, [1, 2, 3], 1, {"a", "b", "c"}, "abc"])
def test_ensure_valid_labware_offset_vector(offset: Dict[str, float]) -> None:
    """It should raise ValueError when given offset is invalid."""
    assert subject.ensure_valid_labware_offset_vector({"x": 1.1, "y": 2, "z": 3.3}) == (
        1.1,
        2,
        3.3,
    )
    with pytest.raises(TypeError):
        subject.ensure_valid_labware_offset_vector(offset)


def test_validate_well_no_location(decoy: Decoy) -> None:
    """Should return a WellTarget with no location."""
    input_location = decoy.mock(cls=Well)
    expected_result = subject.WellTarget(
        well=input_location, location=None, in_place=False
    )

    result = subject.validate_location(location=input_location, last_location=None)

    assert result == expected_result


def test_validate_well_coordinates(decoy: Decoy) -> None:
    """Should return a WellTarget with no location."""
    input_location = Location(point=Point(x=1, y=1, z=2), labware=None)
    expected_result = subject.PointTarget(location=input_location, in_place=False)

    result = subject.validate_location(location=input_location, last_location=None)

    assert result == expected_result


def test_validate_in_place(decoy: Decoy) -> None:
    """Should return an `in_place` PointTarget."""
    input_last_location = Location(point=Point(x=1, y=1, z=2), labware=None)
    expected_result = subject.PointTarget(location=input_last_location, in_place=True)

    result = subject.validate_location(location=None, last_location=input_last_location)

    assert result == expected_result


def test_validate_location_with_well(decoy: Decoy) -> None:
    """Should return a WellTarget with location."""
    mock_well = decoy.mock(cls=Well)
    input_location = Location(point=Point(x=1, y=1, z=1), labware=mock_well)
    expected_result = subject.WellTarget(
        well=mock_well, location=input_location, in_place=False
    )

    result = subject.validate_location(location=input_location, last_location=None)

    assert result == expected_result


def test_validate_last_location(decoy: Decoy) -> None:
    """Should return a WellTarget with location."""
    mock_well = decoy.mock(cls=Well)
    input_last_location = Location(point=Point(x=1, y=1, z=1), labware=mock_well)
    expected_result = subject.WellTarget(
        well=mock_well, location=input_last_location, in_place=True
    )

    result = subject.validate_location(location=None, last_location=input_last_location)

    assert result == expected_result


def test_validate_location_matches_last_location(decoy: Decoy) -> None:
    """Should return an in_place WellTarget."""
    mock_well = decoy.mock(cls=Well)
    input_last_location = Location(point=Point(x=1, y=1, z=1), labware=mock_well)
    input_location = Location(point=Point(x=1, y=1, z=1), labware=mock_well)
    expected_result = subject.WellTarget(
        well=mock_well, location=input_last_location, in_place=True
    )

    result = subject.validate_location(
        location=input_location, last_location=input_last_location
    )

    assert result == expected_result


def test_validate_with_wrong_location_with_last_location() -> None:
    """Should raise a LocationTypeError."""
    with pytest.raises(subject.LocationTypeError):
        subject.validate_location(
            location=42,  # type: ignore[arg-type]
            last_location=Location(point=Point(x=1, y=1, z=1), labware=None),
        )


def test_validate_with_wrong_location() -> None:
    """Should raise a LocationTypeError."""
    with pytest.raises(subject.LocationTypeError):
        subject.validate_location(
            location=42,  # type: ignore[arg-type]
            last_location=None,
        )


def test_validate_raises_no_location_error() -> None:
    """Should raise a NoLocationError."""
    with pytest.raises(subject.NoLocationError):
        subject.validate_location(location=None, last_location=None)


def test_validate_with_labware(decoy: Decoy) -> None:
    """Should return a PointTarget for a non-Well Location."""
    mock_labware = decoy.mock(cls=Labware)
    input_location = Location(point=Point(1, 1, 1), labware=mock_labware)

    result = subject.validate_location(location=input_location, last_location=None)

    assert result == subject.PointTarget(location=input_location, in_place=False)


def test_validate_last_location_with_labware(decoy: Decoy) -> None:
    """Should return a PointTarget for non-Well previous Location."""
    mock_labware = decoy.mock(cls=Labware)
    input_last_location = Location(point=Point(1, 1, 1), labware=mock_labware)

    result = subject.validate_location(location=None, last_location=input_last_location)

    assert result == subject.PointTarget(location=input_last_location, in_place=True)


def test_validate_location_with_trash_bin(decoy: Decoy) -> None:
    """Should return a Trash Bin object."""
    mock_trash = decoy.mock(cls=TrashBin)

    result = subject.validate_location(location=None, last_location=mock_trash)
    assert result.location is mock_trash
    assert result.in_place

    result = subject.validate_location(location=mock_trash, last_location=None)
    assert result.location is mock_trash
    assert not result.in_place

    result = subject.validate_location(location=mock_trash, last_location=mock_trash)
    assert result.location is mock_trash
    assert result.in_place


def test_validate_location_with_waste_chute(decoy: Decoy) -> None:
    """Should return a waste chute object."""
    mock_chute = decoy.mock(cls=WasteChute)

    result = subject.validate_location(location=None, last_location=mock_chute)
    assert result.location is mock_chute
    assert result.in_place

    result = subject.validate_location(location=mock_chute, last_location=None)
    assert result.location is mock_chute
    assert not result.in_place

    result = subject.validate_location(location=mock_chute, last_location=mock_chute)
    assert result.location is mock_chute
    assert result.in_place


def test_ensure_boolean() -> None:
    """It should return a boolean value."""
    assert subject.ensure_boolean(False) is False


@pytest.mark.parametrize("value", [0, "False", "f", 0.0])
def test_ensure_boolean_raises(value: Union[str, int, float]) -> None:
    """It should raise if the value is not a boolean."""
    with pytest.raises(ValueError, match="must be a boolean"):
        subject.ensure_boolean(value)  # type: ignore[arg-type]


@pytest.mark.parametrize("value", [-1.23, -1, 0, 0.0, 1, 1.23])
def test_ensure_float(value: Union[int, float]) -> None:
    """It should return a float value."""
    assert subject.ensure_float(value) == float(value)


def test_ensure_float_raises() -> None:
    """It should raise if the value is not a float or an integer."""
    with pytest.raises(ValueError, match="must be a floating point"):
        subject.ensure_float("1.23")  # type: ignore[arg-type]


@pytest.mark.parametrize("value", [0, 0.1, 1, 1.0])
def test_ensure_positive_float(value: Union[int, float]) -> None:
    """It should return a positive float."""
    assert subject.ensure_positive_float(value) == float(value)


@pytest.mark.parametrize("value", [-1, -1.0, float("inf"), float("-inf"), float("nan")])
def test_ensure_positive_float_raises(value: Union[int, float]) -> None:
    """It should raise if value is not a positive float."""
    with pytest.raises(ValueError, match="(non-infinite|positive float)"):
        subject.ensure_positive_float(value)


def test_ensure_positive_int() -> None:
    """It should return a positive int."""
    assert subject.ensure_positive_int(42) == 42


@pytest.mark.parametrize("value", [1.0, -1.0, -1])
def test_ensure_positive_int_raises(value: Union[int, float]) -> None:
    """It should raise if value is not a positive integer."""
    with pytest.raises(ValueError, match="integer"):
        subject.ensure_positive_int(value)  # type: ignore[arg-type]


def test_validate_coordinates() -> None:
    """It should validate the coordinates and return them as a tuple."""
    assert subject.validate_coordinates([1, 2.0, 3.3]) == (1.0, 2.0, 3.3)


@pytest.mark.parametrize("value", [[1, 2.0], [1, 2.0, 3.3, 4.2], ["1", 2, 3]])
def test_validate_coordinates_raises(value: Sequence[Union[int, float, str]]) -> None:
    """It should raise if value is not a valid sequence of three numbers."""
    with pytest.raises(ValueError, match="(exactly three|must be floats)"):
        subject.validate_coordinates(value)  # type: ignore[arg-type]


@pytest.mark.parametrize(
    argnames=["axis_map", "robot_type", "is_96_channel", "expected_axis_map"],
    argvalues=[
        (
            {"x": 100, "Y": 50, "z_g": 80},
            "OT-3 Standard",
            True,
            {AxisType.X: 100, AxisType.Y: 50, AxisType.Z_G: 80},
        ),
        ({"z_r": 80}, "OT-2 Standard", False, {AxisType.Z_R: 80}),
        (
            {"Z_L": 19, "P_L": 20},
            "OT-2 Standard",
            False,
            {AxisType.Z_L: 19, AxisType.P_L: 20},
        ),
        ({"Q": 5}, "OT-3 Standard", True, {AxisType.Q: 5}),
    ],
)
def test_ensure_axis_map_type_success(
    axis_map: Union[AxisMapType, StringAxisMap],
    robot_type: RobotType,
    is_96_channel: bool,
    expected_axis_map: AxisMapType,
) -> None:
    """Check that axis map type validation returns the correct shape."""
    res = subject.ensure_axis_map_type(axis_map, robot_type, is_96_channel)
    assert res == expected_axis_map


@pytest.mark.parametrize(
    argnames=["axis_map", "robot_type", "is_96_channel", "error_message"],
    argvalues=[
        (
            {AxisType.X: 100, "y": 50},
            "OT-3 Standard",
            True,
            "Please provide an `axis_map` with only string or only AxisType keys",
        ),
        (
            {AxisType.Z_R: 60},
            "OT-3 Standard",
            True,
            "A 96 channel is attached. You cannot move the `Z_R` mount.",
        ),
        (
            {"Z_G": 19, "P_L": 20},
            "OT-2 Standard",
            False,
            "An OT-2 Robot only accepts the following axes ",
        ),
        (
            {"Q": 5},
            "OT-3 Standard",
            False,
            "A 96 channel is not attached. The clamp `Q` motor does not exist.",
        ),
    ],
)
def test_ensure_axis_map_type_failure(
    axis_map: Union[AxisMapType, StringAxisMap],
    robot_type: RobotType,
    is_96_channel: bool,
    error_message: str,
) -> None:
    """Check that axis_map validation occurs for the given scenarios."""
    with pytest.raises(subject.IncorrectAxisError, match=error_message):
        subject.ensure_axis_map_type(axis_map, robot_type, is_96_channel)


@pytest.mark.parametrize(
    argnames=["axis_map", "robot_type", "error_message"],
    argvalues=[
        (
            {AxisType.X: 100, AxisType.P_L: 50},
            "OT-3 Standard",
            "A critical point only accepts Flex gantry axes which are ",
        ),
        (
            {AxisType.Z_G: 60},
            "OT-2 Standard",
            "A critical point only accepts OT-2 gantry axes which are ",
        ),
    ],
)
def test_ensure_only_gantry_axis_map_type(
    axis_map: AxisMapType, robot_type: RobotType, error_message: str
) -> None:
    """Check that gantry axis_map validation occurs for the given scenarios."""
    with pytest.raises(subject.IncorrectAxisError, match=error_message):
        subject.ensure_only_gantry_axis_map_type(axis_map, robot_type)


@pytest.mark.parametrize(
    ["value", "expected_result"],
    [
        ("once", TransferTipPolicyV2.ONCE),
        ("NEVER", TransferTipPolicyV2.NEVER),
        ("alWaYs", TransferTipPolicyV2.ALWAYS),
        ("Per Source", TransferTipPolicyV2.PER_SOURCE),
    ],
)
def test_ensure_new_tip_policy(
    value: str, expected_result: TransferTipPolicyV2
) -> None:
    """It should return the expected tip policy."""
    assert subject.ensure_new_tip_policy(value) == expected_result


def test_ensure_new_tip_policy_raises() -> None:
    """It should raise ValueError for invalid new_tip value."""
    with pytest.raises(ValueError, match="is invalid value for 'new_tip'"):
        subject.ensure_new_tip_policy("blah")


@pytest.mark.parametrize(
    ["target", "expected_raise"],
    [
        (
            "a",
            pytest.raises(
                ValueError, match="'a' is not a valid location for transfer."
            ),
        ),
        (
            ["a"],
            pytest.raises(
                ValueError, match="'a' is not a valid location for transfer."
            ),
        ),
        (
            [("a",)],
            pytest.raises(
                ValueError, match="'a' is not a valid location for transfer."
            ),
        ),
        (
            [],
            pytest.raises(
                ValueError, match="No target well\\(s\\) specified for transfer."
            ),
        ),
    ],
)
def test_ensure_valid_flat_wells_list_raises_for_invalid_targets(
    target: Any,
    expected_raise: ContextManager[Any],
) -> None:
    """It should raise an error if target location is invalid."""
    with expected_raise:
        subject.ensure_valid_flat_wells_list_for_transfer_v2(target)


def test_ensure_valid_flat_wells_list_raises_for_mixed_targets(decoy: Decoy) -> None:
    """It should raise appropriate error if target has mixed valid and invalid wells."""
    target1 = [decoy.mock(cls=Well), "a"]
    with pytest.raises(ValueError, match="'a' is not a valid location for transfer."):
        subject.ensure_valid_flat_wells_list_for_transfer_v2(target1)  # type: ignore[arg-type]

    target2 = [[decoy.mock(cls=Well)], ["a"]]
    with pytest.raises(ValueError, match="'a' is not a valid location for transfer."):
        subject.ensure_valid_flat_wells_list_for_transfer_v2(target2)  # type: ignore[arg-type]


def test_ensure_valid_flat_wells_list(decoy: Decoy) -> None:
    """It should convert the locations to flat lists correctly."""
    target1 = decoy.mock(cls=Well)
    target2 = decoy.mock(cls=Well)

    assert subject.ensure_valid_flat_wells_list_for_transfer_v2(target1) == [target1]
    assert subject.ensure_valid_flat_wells_list_for_transfer_v2([target1, target2]) == [
        target1,
        target2,
    ]
    assert subject.ensure_valid_flat_wells_list_for_transfer_v2(
        [
            [target1, target1],
            [target2, target2],
        ]
    ) == [target1, target1, target2, target2]
    assert subject.ensure_valid_flat_wells_list_for_transfer_v2((target1, target2)) == [
        target1,
        target2,
    ]
    assert subject.ensure_valid_flat_wells_list_for_transfer_v2(
        (
            [target1, target1],
            [target2, target2],
        )
    ) == [target1, target1, target2, target2]


def test_ensure_valid_trash_location_for_transfer_v2(
    decoy: Decoy,
) -> None:
    """It should check that the trash location is valid."""
    mock_well = decoy.mock(cls=Well)
    mock_location = Location(point=Point(x=1, y=1, z=1), labware=mock_well)
    mock_trash_bin = decoy.mock(cls=TrashBin)
    mock_waste_chute = decoy.mock(cls=WasteChute)
    decoy.when(mock_well.top()).then_return(Location(Point(1, 2, 3), labware=mock_well))
    assert subject.ensure_valid_trash_location_for_transfer_v2(mock_well) == Location(
        Point(1, 2, 3), labware=mock_well
    )
    assert (
        subject.ensure_valid_trash_location_for_transfer_v2(mock_location)
        == mock_location
    )
    assert (
        subject.ensure_valid_trash_location_for_transfer_v2(mock_trash_bin)
        == mock_trash_bin
    )
    assert (
        subject.ensure_valid_trash_location_for_transfer_v2(mock_waste_chute)
        == mock_waste_chute
    )


def test_ensure_valid_trash_location_for_transfer_v2_raises(decoy: Decoy) -> None:
    """It should raise an error for invalid trash locations."""
    with pytest.raises(TypeError, match="However, it is '\\['a'\\]'"):
        subject.ensure_valid_trash_location_for_transfer_v2(
            ["a"]  # type: ignore[arg-type]
        )

    mock_labware = decoy.mock(cls=Labware)
    with pytest.raises(TypeError, match=f"However, it is '{mock_labware}'"):
        subject.ensure_valid_trash_location_for_transfer_v2(
            mock_labware  # type: ignore[arg-type]
        )

    with pytest.raises(
        TypeError, match="However, the given location doesn't refer to any well."
    ):
        subject.ensure_valid_trash_location_for_transfer_v2(
            Location(point=Point(x=1, y=1, z=1), labware=None)
        )


def test_is_96_channel() -> None:
    """Iterate through the pipette names and make sure that validation identifies 96 channels."""
    for name in PipetteNameType:
        assert subject.is_pipette_96_channel(name) == ("96" in name.value)



"""Tests for the RunOrchestrator."""

from datetime import datetime
from pathlib import Path
from typing import Generator, Union

import pytest
from decoy import Decoy
from pytest_lazy_fixtures import lf as lazy_fixture

from opentrons import protocol_runner
from opentrons.hardware_control import API as HardwareAPI
from opentrons.hardware_control.modules.types import (
    TemperatureModuleModel as HardwareTemperatureModuleModel,
)
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocol_engine import commands as pe_commands
from opentrons.protocol_engine.error_recovery_policy import ErrorRecoveryPolicy
from opentrons.protocol_engine.errors import RunStoppedError
from opentrons.protocol_engine.resources.camera_provider import CameraProvider
from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.types import (
    ModuleModel as EngineModuleModel,
)
from opentrons.protocol_engine.types import (
    PostRunHardwareState,
)
from opentrons.protocol_reader import (
    JsonProtocolConfig,
    ProtocolSource,
    PythonProtocolConfig,
)
from opentrons.protocol_runner.protocol_runner import (
    JsonRunner,
    LiveRunner,
    PythonAndLegacyRunner,
)
from opentrons.protocol_runner.run_orchestrator import (
    ParseMode,
    RunNotFound,
    RunOrchestrator,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.parse import PythonParseMode


@pytest.fixture
def mock_protocol_python_runner(decoy: Decoy) -> PythonAndLegacyRunner:
    """Get a mocked out PythonAndLegacyRunner dependency."""
    return decoy.mock(cls=PythonAndLegacyRunner)


@pytest.fixture
def mock_protocol_json_runner(decoy: Decoy) -> JsonRunner:
    """Get a mocked out PythonAndLegacyRunner dependency."""
    return decoy.mock(cls=JsonRunner)


@pytest.fixture
def mock_setup_runner(decoy: Decoy) -> LiveRunner:
    """Get a mocked out LiveRunner dependency."""
    return decoy.mock(cls=LiveRunner)


@pytest.fixture
def mock_fixit_runner(decoy: Decoy) -> LiveRunner:
    """Get a mocked out LiveRunner dependency."""
    return decoy.mock(cls=LiveRunner)


@pytest.fixture
def mock_protocol_live_runner(decoy: Decoy) -> LiveRunner:
    """Get a mock of a LiveRunner for protocol commands."""
    return decoy.mock(cls=LiveRunner)


@pytest.fixture
def mock_protocol_engine(decoy: Decoy) -> ProtocolEngine:
    """Get a mocked out ProtocolEngine dependency."""
    return decoy.mock(cls=ProtocolEngine)


@pytest.fixture
def mock_hardware_api(decoy: Decoy) -> HardwareAPI:
    """Get a mocked out HardwareAPI dependency."""
    return decoy.mock(cls=HardwareAPI)


@pytest.fixture
def mock_camera_provider(decoy: Decoy) -> CameraProvider:
    """Get a mocked out Camera Provider dependency."""
    return decoy.mock(cls=CameraProvider)


@pytest.fixture
def json_protocol_subject(
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    mock_protocol_json_runner: JsonRunner,
    mock_fixit_runner: LiveRunner,
    mock_setup_runner: LiveRunner,
    mock_protocol_live_runner: LiveRunner,
    mock_camera_provider: CameraProvider,
) -> RunOrchestrator:
    """Get a RunOrchestrator subject with a json runner."""
    return RunOrchestrator(
        protocol_engine=mock_protocol_engine,
        hardware_api=mock_hardware_api,
        fixit_runner=mock_fixit_runner,
        setup_runner=mock_setup_runner,
        json_or_python_protocol_runner=mock_protocol_json_runner,
        protocol_live_runner=mock_protocol_live_runner,
        camera_provider=mock_camera_provider,
    )


@pytest.fixture
def python_protocol_subject(
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    mock_protocol_python_runner: PythonAndLegacyRunner,
    mock_fixit_runner: LiveRunner,
    mock_setup_runner: LiveRunner,
    mock_protocol_live_runner: LiveRunner,
    mock_camera_provider: CameraProvider,
) -> RunOrchestrator:
    """Get a RunOrchestrator subject with a python runner."""
    return RunOrchestrator(
        protocol_engine=mock_protocol_engine,
        hardware_api=mock_hardware_api,
        fixit_runner=mock_fixit_runner,
        setup_runner=mock_setup_runner,
        json_or_python_protocol_runner=mock_protocol_python_runner,
        protocol_live_runner=mock_protocol_live_runner,
        camera_provider=mock_camera_provider,
    )


@pytest.fixture
def live_protocol_subject(
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    mock_fixit_runner: LiveRunner,
    mock_setup_runner: LiveRunner,
    mock_protocol_live_runner: LiveRunner,
    mock_camera_provider: CameraProvider,
) -> RunOrchestrator:
    """Get a RunOrchestrator subject with a live runner."""
    return RunOrchestrator(
        protocol_engine=mock_protocol_engine,
        hardware_api=mock_hardware_api,
        fixit_runner=mock_fixit_runner,
        setup_runner=mock_setup_runner,
        protocol_live_runner=mock_protocol_live_runner,
        camera_provider=mock_camera_provider,
    )


@pytest.mark.parametrize(
    "input_protocol_config, mock_protocol_runner, subject",
    [
        (
            JsonProtocolConfig(schema_version=7),
            lazy_fixture("mock_protocol_json_runner"),
            lazy_fixture("json_protocol_subject"),
        ),
        (
            PythonProtocolConfig(api_version=APIVersion(2, 14)),
            lazy_fixture("mock_protocol_python_runner"),
            lazy_fixture("python_protocol_subject"),
        ),
    ],
)
def test_build_run_orchestrator_provider(
    decoy: Decoy,
    monkeypatch: pytest.MonkeyPatch,
    subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    input_protocol_config: Union[PythonProtocolConfig, JsonProtocolConfig],
    mock_setup_runner: LiveRunner,
    mock_fixit_runner: LiveRunner,
    mock_protocol_runner: Union[PythonAndLegacyRunner, JsonRunner],
    mock_camera_provider: CameraProvider,
) -> None:
    """Should get a RunOrchestrator instance."""
    mock_create_runner_func = decoy.mock(func=protocol_runner.create_protocol_runner)
    monkeypatch.setattr(
        protocol_runner, "create_protocol_runner", mock_create_runner_func
    )

    decoy.when(
        mock_create_runner_func(
            protocol_config=input_protocol_config,
            protocol_engine=mock_protocol_engine,
            hardware_api=mock_hardware_api,
            post_run_hardware_state=PostRunHardwareState.HOME_AND_STAY_ENGAGED,
            drop_tips_after_run=True,
        )
    ).then_return(mock_protocol_runner)

    result = subject.build_orchestrator(
        protocol_engine=mock_protocol_engine,
        camera_provider=mock_camera_provider,
        hardware_api=mock_hardware_api,
        protocol_config=input_protocol_config,
    )

    assert isinstance(result, RunOrchestrator)
    assert isinstance(result._setup_runner, LiveRunner)
    assert isinstance(result._fixit_runner, LiveRunner)
    assert isinstance(result._protocol_runner, (PythonAndLegacyRunner, JsonRunner))


@pytest.mark.parametrize(
    "subject, runner",
    [
        (
            lazy_fixture("json_protocol_subject"),
            lazy_fixture("mock_protocol_json_runner"),
        ),
        (
            lazy_fixture("python_protocol_subject"),
            lazy_fixture("mock_protocol_python_runner"),
        ),
    ],
)
async def test_run_calls_protocol_runner(
    subject: RunOrchestrator,
    runner: Union[JsonRunner, PythonAndLegacyRunner],
    decoy: Decoy,
) -> None:
    """Should call protocol runner run method."""
    await subject.run(deck_configuration=[])
    decoy.verify(
        await runner.run(
            deck_configuration=[], run_time_param_values=None, protocol_source=None
        )
    )


async def test_run_calls_protocol_live_runner(
    live_protocol_subject: RunOrchestrator,
    mock_protocol_live_runner: LiveRunner,
    decoy: Decoy,
) -> None:
    """Should call protocol runner run method."""
    await live_protocol_subject.run(deck_configuration=[])
    decoy.verify(await mock_protocol_live_runner.run(deck_configuration=[]))


def test_get_run_time_parameters_returns_an_empty_list_no_protocol(
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should return an empty list in case the protocol runner is not initialized."""
    result = live_protocol_subject.get_run_time_parameters()
    assert result == []


def test_get_run_time_parameters_returns_an_empty_list_json_runner(
    decoy: Decoy,
    mock_protocol_json_runner: JsonRunner,
    json_protocol_subject: RunOrchestrator,
) -> None:
    """Should return an empty list in case the protocol runner is a json runner."""
    decoy.when(mock_protocol_json_runner.run_time_parameters).then_return([])
    result = json_protocol_subject.get_run_time_parameters()
    assert result == []


@pytest.mark.parametrize(
    "wait_for_interval_input, verify_calls", [(True, 1), (False, 0)]
)
async def test_add_command_and_wait_for_interval(
    decoy: Decoy,
    json_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
    wait_for_interval_input: bool,
    verify_calls: int,
) -> None:
    """Should add a command a wait for it to complete."""
    load_command = pe_commands.HomeCreate.model_construct(
        params=pe_commands.HomeParams.model_construct()
    )
    added_command = pe_commands.Home(
        params=pe_commands.HomeParams.model_construct(),
        id="test-123",
        createdAt=datetime(year=2024, month=1, day=1),
        key="123",
        status=pe_commands.CommandStatus.QUEUED,
    )
    decoy.when(
        mock_protocol_engine.add_command(request=load_command, failed_command_id=None)
    ).then_return(added_command)

    result = await json_protocol_subject.add_command_and_wait_for_interval(
        command=load_command, wait_until_complete=wait_for_interval_input, timeout=999
    )

    assert result == added_command

    decoy.verify(
        await mock_protocol_engine.wait_for_command(command_id="test-123"),
        times=verify_calls,
    )


def test_estop(
    decoy: Decoy,
    live_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
) -> None:
    """Verify an estop call."""
    live_protocol_subject.estop()
    decoy.verify(mock_protocol_engine.estop())


async def test_use_attached_modules(
    decoy: Decoy,
    live_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
) -> None:
    """Verify a call to use_attached_modules."""
    await live_protocol_subject.use_attached_modules(modules_by_id={})
    decoy.verify(await mock_protocol_engine.use_attached_modules({}))


def test_get_protocol_runner(
    json_protocol_subject: RunOrchestrator,
    python_protocol_subject: RunOrchestrator,
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should return the equivalent runner."""
    json_runner = json_protocol_subject.get_protocol_runner()
    assert isinstance(json_runner, JsonRunner)

    python_runner = python_protocol_subject.get_protocol_runner()
    assert isinstance(python_runner, PythonAndLegacyRunner)

    live_runner = live_protocol_subject.get_protocol_runner()
    assert live_runner is None


async def test_load_json(
    decoy: Decoy,
    json_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
    mock_protocol_json_runner: JsonRunner,
) -> None:
    """Should load a json protocol runner."""
    protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=6),
        content_hash="abc123",
    )
    await json_protocol_subject.load(
        protocol_source=protocol_source,
        run_time_param_values=None,
        run_time_param_paths=None,
        parse_mode=ParseMode.NORMAL,
    )

    decoy.verify(await mock_protocol_json_runner.load(protocol_source))


async def test_load_python(
    decoy: Decoy,
    python_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
    mock_protocol_python_runner: PythonAndLegacyRunner,
) -> None:
    """Should load a json protocol runner."""
    protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=6),
        content_hash="abc123",
    )
    await python_protocol_subject.load(
        protocol_source=protocol_source,
        parse_mode=ParseMode.NORMAL,
        run_time_param_values=None,
        run_time_param_paths=None,
    )

    decoy.verify(
        await mock_protocol_python_runner.load(
            protocol_source=protocol_source,
            python_parse_mode=PythonParseMode.NORMAL,
            run_time_param_values=None,
            run_time_param_paths=None,
        )
    )


async def test_load_json_raises_no_protocol(
    decoy: Decoy,
    live_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
) -> None:
    """Should raise that there is no protocol runner."""
    protocol_source = ProtocolSource(
        directory=Path("/dev/null"),
        main_file=Path("/dev/null/abc.json"),
        files=[],
        metadata={},
        robot_type="OT-2 Standard",
        config=JsonProtocolConfig(schema_version=6),
        content_hash="abc123",
    )
    with pytest.raises(AssertionError):
        await live_protocol_subject.load(
            protocol_source=protocol_source,
            run_time_param_values=None,
            run_time_param_paths=None,
            parse_mode=ParseMode.NORMAL,
        )


def test_get_run_id(
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    mock_fixit_runner: LiveRunner,
    mock_setup_runner: LiveRunner,
    mock_protocol_live_runner: LiveRunner,
    mock_camera_provider: CameraProvider,
) -> None:
    """Should get run_id if builder was created with a run id."""
    orchestrator = RunOrchestrator(
        run_id="test-123",
        protocol_engine=mock_protocol_engine,
        hardware_api=mock_hardware_api,
        fixit_runner=mock_fixit_runner,
        setup_runner=mock_setup_runner,
        protocol_live_runner=mock_protocol_live_runner,
        camera_provider=mock_camera_provider,
    )
    assert orchestrator.run_id == "test-123"


def test_get_run_id_raises(
    mock_protocol_engine: ProtocolEngine,
    mock_hardware_api: HardwareAPI,
    mock_fixit_runner: LiveRunner,
    mock_setup_runner: LiveRunner,
    mock_protocol_live_runner: LiveRunner,
    mock_camera_provider: CameraProvider,
) -> None:
    """Should get run_id if builder was created with a run id."""
    orchestrator = RunOrchestrator(
        protocol_engine=mock_protocol_engine,
        hardware_api=mock_hardware_api,
        fixit_runner=mock_fixit_runner,
        setup_runner=mock_setup_runner,
        protocol_live_runner=mock_protocol_live_runner,
        camera_provider=mock_camera_provider,
    )
    with pytest.raises(RunNotFound):
        orchestrator.run_id


def test_get_is_okay_to_clear(
    decoy: Decoy,
    mock_protocol_engine: ProtocolEngine,
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should return if is ok to clear run or not."""
    decoy.when(
        mock_protocol_engine.state_view.commands.get_is_okay_to_clear()
    ).then_return(True)
    result = live_protocol_subject.get_is_okay_to_clear()

    assert result is True

    decoy.when(
        mock_protocol_engine.state_view.commands.get_is_okay_to_clear()
    ).then_return(False)
    result = live_protocol_subject.get_is_okay_to_clear()

    assert result is False


def test_prepare(
    decoy: Decoy,
    live_protocol_subject: RunOrchestrator,
    mock_protocol_live_runner: LiveRunner,
) -> None:
    """Verify prepare calls runner prepare."""
    live_protocol_subject.prepare()
    decoy.verify(mock_protocol_live_runner.prepare())


async def test_stop(
    decoy: Decoy,
    mock_protocol_engine: ProtocolEngine,
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should verify a call to stop/finish the run."""
    decoy.when(mock_protocol_engine.state_view.commands.has_been_played()).then_return(
        True
    )
    await live_protocol_subject.stop()
    decoy.verify(await mock_protocol_engine.request_stop())

    decoy.when(mock_protocol_engine.state_view.commands.has_been_played()).then_return(
        False
    )
    await live_protocol_subject.stop()
    decoy.verify(
        await mock_protocol_engine.finish(
            error=None,
            drop_tips_after_run=False,
            set_run_status=False,
            post_run_hardware_state=PostRunHardwareState.STAY_ENGAGED_IN_PLACE,
        )
    )


async def test_command_generator(
    decoy: Decoy,
    mock_protocol_engine: ProtocolEngine,
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should get the next command to execute."""

    def get_next_to_execute() -> Generator[str, None, None]:
        yield "command-id-1"
        yield "command-id-2"
        raise RunStoppedError()

    get_next_to_execute_results = get_next_to_execute()

    mock_state_store = decoy.mock(cls=StateStore)
    decoy.when(mock_protocol_engine._state_store).then_return(mock_state_store)

    decoy.when(
        await mock_protocol_engine._state_store.wait_for(
            condition=mock_protocol_engine.state_view.commands.get_next_to_execute
        )
    ).then_do(lambda *args, **kwargs: next(get_next_to_execute_results))

    index = 1
    async for command in live_protocol_subject.command_generator():
        assert command == f"command-id-{index}"
        index = index + 1


def test_create_error_recovery_policy(
    decoy: Decoy,
    mock_protocol_engine: ProtocolEngine,
    live_protocol_subject: RunOrchestrator,
) -> None:
    """Should call PE set_error_recovery_policy."""
    policy = decoy.mock(cls=ErrorRecoveryPolicy)
    live_protocol_subject.set_error_recovery_policy(policy)
    decoy.verify(mock_protocol_engine.set_error_recovery_policy(policy))


async def test_transform_module_model_in_ame(
    decoy: Decoy,
    python_protocol_subject: RunOrchestrator,
    mock_protocol_engine: ProtocolEngine,
) -> None:
    """It should make the module model the PE kind in its asynch module error call."""
    decoy.when(
        await mock_protocol_engine.async_module_error(
            module_model=EngineModuleModel.TEMPERATURE_MODULE_V2,
            serial="whatever",
        )
    ).then_return(True)
    assert (
        await python_protocol_subject.asynchronous_module_error(
            module_model=HardwareTemperatureModuleModel.TEMPERATURE_V2,
            module_serial="whatever",
        )
        is True
    )



"""Addressable area state view tests.

DEPRECATED: Testing AddressableAreaView independently of AddressableAreaStore is no
longer helpful. Try to add new tests to test_addressable_area_state.py, where they can be
tested together, treating AddressableAreaState as a private implementation detail.
"""

import inspect
from typing import Dict, Optional, Set, cast
from unittest.mock import sentinel

import pytest
from decoy import Decoy

from opentrons_shared_data.deck.types import DeckDefinitionV5
from opentrons_shared_data.labware.types import LocatingFeatures
from opentrons_shared_data.module.types import ModuleOrientation
from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_engine.errors import (
    AddressableAreaDoesNotExistError,
    AreaNotInDeckConfigurationError,
    IncompatibleAddressableAreaError,
    SlotDoesNotExistError,
)
from opentrons.protocol_engine.resources import deck_configuration_provider
from opentrons.protocol_engine.state.addressable_areas import (
    AddressableAreaState,
    AddressableAreaView,
)
from opentrons.protocol_engine.types import (
    AddressableArea,
    AddressableOffsetVector,
    AreaType,
    DeckConfigurationType,
    DeckPoint,
    Dimensions,
    PotentialCutoutFixture,
)
from opentrons.types import DeckSlotName, Point


@pytest.fixture(autouse=True)
def patch_mock_deck_configuration_provider(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out deck_configuration_provider.py functions."""
    for name, func in inspect.getmembers(
        deck_configuration_provider, inspect.isfunction
    ):
        monkeypatch.setattr(deck_configuration_provider, name, decoy.mock(func=func))


def get_addressable_area_view(
    loaded_addressable_areas_by_name: Optional[Dict[str, AddressableArea]] = None,
    potential_cutout_fixtures_by_cutout_id: Optional[
        Dict[str, Set[PotentialCutoutFixture]]
    ] = None,
    deck_definition: Optional[DeckDefinitionV5] = None,
    deck_configuration: Optional[DeckConfigurationType] = None,
    robot_type: RobotType = "OT-3 Standard",
    use_simulated_deck_config: bool = False,
) -> AddressableAreaView:
    """Get a labware view test subject."""
    state = AddressableAreaState(
        loaded_addressable_areas_by_name=loaded_addressable_areas_by_name or {},
        potential_cutout_fixtures_by_cutout_id=potential_cutout_fixtures_by_cutout_id
        or {},
        deck_definition=deck_definition or cast(DeckDefinitionV5, {"otId": "fake"}),
        robot_definition={
            "displayName": "OT-3",
            "robotType": "OT-3 Standard",
            "models": ["OT-3 Standard"],
            "extents": [477.2, 493.8, 0.0],
            "paddingOffsets": {
                "rear": -177.42,
                "front": 51.8,
                "leftSide": 31.88,
                "rightSide": -80.32,
            },
            "mountOffsets": {
                "left": [-13.5, -60.5, 255.675],
                "right": [40.5, -60.5, 255.675],
                "gripper": [84.55, -12.75, 93.85],
            },
        },
        deck_configuration=deck_configuration or [],
        robot_type=robot_type,
        use_simulated_deck_config=use_simulated_deck_config,
    )

    return AddressableAreaView(state=state)


def test_get_all_cutout_fixtures_simulated_deck_config() -> None:
    """It should return no cutout fixtures when the deck config is simulated."""
    subject = get_addressable_area_view(
        deck_configuration=None,
        use_simulated_deck_config=True,
    )
    assert subject.get_all_cutout_fixtures() is None


def test_get_all_cutout_fixtures_non_simulated_deck_config() -> None:
    """It should return the cutout fixtures from the deck config, if it's not simulated."""
    subject = get_addressable_area_view(
        deck_configuration=[
            ("cutout-id-1", "cutout-fixture-id-1", None),
            ("cutout-id-2", "cutout-fixture-id-2", None),
        ],
        use_simulated_deck_config=False,
    )
    assert subject.get_all_cutout_fixtures() == [
        "cutout-fixture-id-1",
        "cutout-fixture-id-2",
    ]


def test_get_loaded_addressable_area() -> None:
    """It should get the loaded addressable area."""
    addressable_area = AddressableArea(
        area_name="area",
        area_type=AreaType.SLOT,
        base_slot=DeckSlotName.SLOT_D3,
        display_name="fancy name",
        bounding_box=Dimensions(x=1, y=2, z=3),
        position=AddressableOffsetVector(x=7, y=8, z=9),
        features=LocatingFeatures(),
        compatible_module_types=["magneticModuleType"],
        mating_surface_unit_vector=[-1, 1, -1],
        orientation=ModuleOrientation.NOT_APPLICABLE,
    )
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={"abc": addressable_area}
    )

    assert subject.get_addressable_area("abc") is addressable_area


def test_get_loaded_addressable_area_raises() -> None:
    """It should raise if the addressable area does not exist."""
    subject = get_addressable_area_view()

    with pytest.raises(AreaNotInDeckConfigurationError):
        subject.get_addressable_area("abc")


def test_get_addressable_area_for_simulation_already_loaded() -> None:
    """It should get the addressable area for a simulation that has not been loaded yet."""
    addressable_area = AddressableArea(
        area_name="area",
        area_type=AreaType.SLOT,
        base_slot=DeckSlotName.SLOT_D3,
        display_name="fancy name",
        bounding_box=Dimensions(x=1, y=2, z=3),
        position=AddressableOffsetVector(x=7, y=8, z=9),
        compatible_module_types=["magneticModuleType"],
        features=LocatingFeatures(),
        mating_surface_unit_vector=[-1, 1, -1],
        orientation=ModuleOrientation.NOT_APPLICABLE,
    )
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={"abc": addressable_area},
        use_simulated_deck_config=True,
    )

    assert subject.get_addressable_area("abc") is addressable_area


def test_get_addressable_area_for_simulation_not_loaded(decoy: Decoy) -> None:
    """It should get the addressable area for a simulation that has not been loaded yet."""
    subject = get_addressable_area_view(
        potential_cutout_fixtures_by_cutout_id={
            "cutoutA1": {
                PotentialCutoutFixture(
                    cutout_id="cutoutA1",
                    cutout_fixture_id="blah",
                    provided_addressable_areas=frozenset(),
                )
            }
        },
        use_simulated_deck_config=True,
        deck_definition=sentinel.deck_definition,
    )

    addressable_area = AddressableArea(
        area_name="area",
        area_type=AreaType.SLOT,
        base_slot=DeckSlotName.SLOT_D3,
        display_name="fancy name",
        bounding_box=Dimensions(x=1, y=2, z=3),
        position=AddressableOffsetVector(x=7, y=8, z=9),
        compatible_module_types=["magneticModuleType"],
        features=LocatingFeatures(),
        mating_surface_unit_vector=[-1, 1, -1],
        orientation=ModuleOrientation.NOT_APPLICABLE,
    )

    decoy.when(
        deck_configuration_provider.get_potential_cutout_fixtures(
            "abc",
            sentinel.deck_definition,
        )
    ).then_return(
        (
            "cutoutA1",
            {
                PotentialCutoutFixture(
                    cutout_id="cutoutA1",
                    cutout_fixture_id="blah",
                    provided_addressable_areas=frozenset(),
                )
            },
        )
    )

    decoy.when(
        deck_configuration_provider.get_cutout_position(
            "cutoutA1",
            sentinel.deck_definition,
        )
    ).then_return(DeckPoint(x=1, y=2, z=3))

    decoy.when(
        deck_configuration_provider.get_addressable_area_from_name(
            "abc",
            DeckPoint(x=1, y=2, z=3),
            sentinel.deck_definition,
        )
    ).then_return(addressable_area)

    assert subject.get_addressable_area("abc") is addressable_area


def test_get_addressable_area_for_simulation_raises(decoy: Decoy) -> None:
    """It should raise if the requested addressable area is incompatible with loaded ones."""
    subject = get_addressable_area_view(
        potential_cutout_fixtures_by_cutout_id={
            "123": {
                PotentialCutoutFixture(
                    cutout_id="789",
                    cutout_fixture_id="bleh",
                    provided_addressable_areas=frozenset(),
                )
            }
        },
        use_simulated_deck_config=True,
        deck_definition=sentinel.deck_definition,
    )

    decoy.when(
        deck_configuration_provider.get_potential_cutout_fixtures(
            "abc", sentinel.deck_definition
        )
    ).then_return(
        (
            "123",
            {
                PotentialCutoutFixture(
                    cutout_id="123",
                    cutout_fixture_id="blah",
                    provided_addressable_areas=frozenset(),
                )
            },
        )
    )

    decoy.when(
        deck_configuration_provider.get_provided_addressable_area_names(
            "bleh", "789", sentinel.deck_definition
        )
    ).then_return([])

    with pytest.raises(IncompatibleAddressableAreaError):
        subject.get_addressable_area("abc")


def test_get_addressable_area_position() -> None:
    """It should get the absolute location of the addressable area."""
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={
            "abc": AddressableArea(
                area_name="area",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_D3,
                display_name="fancy name",
                bounding_box=Dimensions(x=10, y=20, z=30),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                compatible_module_types=[],
                features=LocatingFeatures(),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            )
        }
    )

    result = subject.get_addressable_area_position("abc")
    assert result == Point(1, 2, 3)


def test_get_addressable_area_move_to_location() -> None:
    """It should get the absolute location of an addressable area's move to location."""
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={
            "abc": AddressableArea(
                area_name="area",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_D3,
                display_name="fancy name",
                bounding_box=Dimensions(x=10, y=20, z=30),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                compatible_module_types=[],
                features=LocatingFeatures(),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            )
        }
    )

    result = subject.get_addressable_area_move_to_location("abc")
    assert result == Point(6, 12, 33)


def test_get_addressable_area_center() -> None:
    """It should get the absolute location of an addressable area's center."""
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={
            "abc": AddressableArea(
                area_name="area",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_D3,
                display_name="fancy name",
                bounding_box=Dimensions(x=10, y=20, z=30),
                position=AddressableOffsetVector(x=1, y=2, z=3),
                compatible_module_types=[],
                features=LocatingFeatures(),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            )
        }
    )

    result = subject.get_addressable_area_center("abc")
    assert result == Point(6, 12, 3)


def test_get_fixture_height(decoy: Decoy) -> None:
    """It should return the height of the requested fixture."""
    subject = get_addressable_area_view(deck_definition=sentinel.deck_definition)
    decoy.when(
        deck_configuration_provider.get_cutout_fixture(
            "someShortCutoutFixture", sentinel.deck_definition
        )
    ).then_return(
        {
            "height": 10,
            # These values don't matter:
            "id": "id",
            "expectOpentronsModuleSerialNumber": False,
            "fixtureGroup": {},
            "mayMountTo": [],
            "displayName": "",
            "providesAddressableAreas": {},
        }
    )

    decoy.when(
        deck_configuration_provider.get_cutout_fixture(
            "someTallCutoutFixture", sentinel.deck_definition
        )
    ).then_return(
        {
            "height": 9000.1,
            # These values don't matter:
            "id": "id",
            "expectOpentronsModuleSerialNumber": False,
            "fixtureGroup": {},
            "mayMountTo": [],
            "displayName": "",
            "providesAddressableAreas": {},
        }
    )

    assert subject.get_fixture_height("someShortCutoutFixture") == 10
    assert subject.get_fixture_height("someTallCutoutFixture") == 9000.1


def test_get_slot_definition() -> None:
    """It should return a deck slot's definition."""
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={
            "6": AddressableArea(
                area_name="area",
                area_type=AreaType.SLOT,
                base_slot=DeckSlotName.SLOT_D3,
                display_name="fancy name",
                bounding_box=Dimensions(x=1, y=2, z=3),
                position=AddressableOffsetVector(x=7, y=8, z=9),
                compatible_module_types=["magneticModuleType"],
                features=LocatingFeatures(),
                mating_surface_unit_vector=[-1, 1, -1],
                orientation=ModuleOrientation.NOT_APPLICABLE,
            )
        }
    )

    result = subject.get_slot_definition(DeckSlotName.SLOT_6.id)

    assert result == {
        "id": "area",
        "position": [7, 8, 9],
        "boundingBox": {
            "xDimension": 1,
            "yDimension": 2,
            "zDimension": 3,
        },
        "displayName": "fancy name",
        "compatibleModuleTypes": ["magneticModuleType"],
        "features": {},
        "orientation": ModuleOrientation.NOT_APPLICABLE,
    }


def test_get_slot_definition_raises_with_bad_slot_name(decoy: Decoy) -> None:
    """It should raise a SlotDoesNotExistError if a bad slot name is given."""
    deck_definition = cast(DeckDefinitionV5, {"otId": "fake"})
    subject = get_addressable_area_view(
        deck_definition=deck_definition,
    )

    decoy.when(
        deck_configuration_provider.get_potential_cutout_fixtures(
            "foo", deck_definition
        )
    ).then_raise(AddressableAreaDoesNotExistError())

    with pytest.raises(SlotDoesNotExistError):
        subject.get_slot_definition("foo")


def test_raise_if_area_not_in_deck_configuration_on_robot(decoy: Decoy) -> None:
    """It should raise if the requested addressable area name is not loaded in state."""
    subject = get_addressable_area_view(
        loaded_addressable_areas_by_name={"real": decoy.mock(cls=AddressableArea)}
    )

    subject.raise_if_area_not_in_deck_configuration("real")

    with pytest.raises(AreaNotInDeckConfigurationError):
        subject.raise_if_area_not_in_deck_configuration("fake")


def test_raise_if_area_not_in_deck_configuration_simulated_config(decoy: Decoy) -> None:
    """It should raise if the requested addressable area name is not loaded in state."""
    subject = get_addressable_area_view(
        use_simulated_deck_config=True,
        potential_cutout_fixtures_by_cutout_id={
            "waluigi": {
                PotentialCutoutFixture(
                    cutout_id="fire flower",
                    cutout_fixture_id="1up",
                    provided_addressable_areas=frozenset(),
                )
            },
            "wario": {
                PotentialCutoutFixture(
                    cutout_id="mushroom",
                    cutout_fixture_id="star",
                    provided_addressable_areas=frozenset(),
                )
            },
        },
        deck_definition=sentinel.deck_definition,
    )

    decoy.when(
        deck_configuration_provider.get_potential_cutout_fixtures(
            "mario", sentinel.deck_definition
        )
    ).then_return(
        (
            "wario",
            {
                PotentialCutoutFixture(
                    cutout_id="mushroom",
                    cutout_fixture_id="star",
                    provided_addressable_areas=frozenset(),
                )
            },
        )
    )

    subject.raise_if_area_not_in_deck_configuration("mario")

    decoy.when(
        deck_configuration_provider.get_potential_cutout_fixtures(
            "luigi", sentinel.deck_definition
        )
    ).then_return(
        (
            "waluigi",
            {
                PotentialCutoutFixture(
                    cutout_id="mushroom",
                    cutout_fixture_id="star",
                    provided_addressable_areas=frozenset(),
                )
            },
        )
    )

    decoy.when(
        deck_configuration_provider.get_provided_addressable_area_names(
            "1up", "fire flower", sentinel.deck_definition
        )
    ).then_return([])

    decoy.when(
        deck_configuration_provider.get_addressable_area_display_name(
            "luigi", sentinel.deck_definition
        )
    ).then_return("super luigi")

    with pytest.raises(IncompatibleAddressableAreaError):
        subject.raise_if_area_not_in_deck_configuration("luigi")



"""Test load-liquid command."""

import pytest
from decoy import Decoy

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    LiquidClassSchemaV1,
)

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.load_liquid_class import (
    LoadLiquidClassImplementation,
    LoadLiquidClassParams,
    LoadLiquidClassResult,
)
from opentrons.protocol_engine.errors import (
    LiquidClassDoesNotExistError,
    LiquidClassRedefinitionError,
)
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LiquidClassRecord


@pytest.fixture
def liquid_class_record(
    minimal_liquid_class_def2: LiquidClassSchemaV1,
) -> LiquidClassRecord:
    """A dummy LiquidClassRecord for testing."""
    pipette_0 = minimal_liquid_class_def2.byPipette[0]
    by_tip_type_0 = pipette_0.byTipType[0]
    return LiquidClassRecord(
        liquidClassName=minimal_liquid_class_def2.liquidClassName,
        pipetteModel=pipette_0.pipetteModel,
        tiprack=by_tip_type_0.tiprack,
        aspirate=by_tip_type_0.aspirate,
        singleDispense=by_tip_type_0.singleDispense,
        multiDispense=by_tip_type_0.multiDispense,
    )


async def test_load_liquid_class_new_liquid_class_no_id(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    liquid_class_record: LiquidClassRecord,
) -> None:
    """Load a new liquid class with no liquidClassId specified. Should assign a new ID."""
    subject = LoadLiquidClassImplementation(
        state_view=state_view, model_utils=model_utils
    )
    decoy.when(model_utils.generate_id()).then_return("new-generated-id")

    params = LoadLiquidClassParams(liquidClassRecord=liquid_class_record)
    result = await subject.execute(params)
    assert result == SuccessData(
        public=LoadLiquidClassResult(liquidClassId="new-generated-id"),
        state_update=update_types.StateUpdate(
            liquid_class_loaded=update_types.LiquidClassLoadedUpdate(
                liquid_class_id="new-generated-id",
                liquid_class_record=liquid_class_record,
            )
        ),
    )


async def test_load_liquid_class_existing_liquid_class_no_id(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    liquid_class_record: LiquidClassRecord,
) -> None:
    """Load an existing liquid class with no liquidClassId specified. Should find and reuse existing ID."""
    subject = LoadLiquidClassImplementation(
        state_view=state_view, model_utils=model_utils
    )
    decoy.when(
        state_view.liquid_classes.get_id_for_liquid_class_record(liquid_class_record)
    ).then_return("existing-id")

    params = LoadLiquidClassParams(liquidClassRecord=liquid_class_record)
    result = await subject.execute(params)
    assert (
        result
        == SuccessData(
            public=LoadLiquidClassResult(liquidClassId="existing-id"),
            state_update=update_types.StateUpdate(),  # no state change since liquid class already loaded
        )
    )


async def test_load_liquid_class_new_liquid_class_specified_id(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    liquid_class_record: LiquidClassRecord,
) -> None:
    """Load a new liquid class with the liquidClassId specified. Should store the new liquid class."""
    subject = LoadLiquidClassImplementation(
        state_view=state_view, model_utils=model_utils
    )
    decoy.when(state_view.liquid_classes.get("liquid-class-1")).then_raise(
        LiquidClassDoesNotExistError()
    )

    params = LoadLiquidClassParams(
        liquidClassId="liquid-class-1", liquidClassRecord=liquid_class_record
    )
    result = await subject.execute(params)
    assert result == SuccessData(
        public=LoadLiquidClassResult(liquidClassId="liquid-class-1"),
        state_update=update_types.StateUpdate(
            liquid_class_loaded=update_types.LiquidClassLoadedUpdate(
                liquid_class_id="liquid-class-1",
                liquid_class_record=liquid_class_record,
            )
        ),
    )


async def test_load_liquid_class_existing_liquid_class_specified_id(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    liquid_class_record: LiquidClassRecord,
) -> None:
    """Load a liquid class with a liquidClassId that was already loaded before. Should be a no-op."""
    subject = LoadLiquidClassImplementation(
        state_view=state_view, model_utils=model_utils
    )
    decoy.when(state_view.liquid_classes.get("liquid-class-1")).then_return(
        liquid_class_record
    )

    params = LoadLiquidClassParams(
        liquidClassId="liquid-class-1", liquidClassRecord=liquid_class_record
    )
    result = await subject.execute(params)
    assert (
        result
        == SuccessData(
            public=LoadLiquidClassResult(liquidClassId="liquid-class-1"),
            state_update=update_types.StateUpdate(),  # no state change since liquid class already loaded
        )
    )


async def test_load_liquid_class_conflicting_definition_for_id(
    decoy: Decoy,
    state_view: StateView,
    model_utils: ModelUtils,
    liquid_class_record: LiquidClassRecord,
) -> None:
    """Should raise when we try to load a modified liquid class with the same liquidClassId."""
    subject = LoadLiquidClassImplementation(
        state_view=state_view, model_utils=model_utils
    )
    decoy.when(state_view.liquid_classes.get("liquid-class-1")).then_return(
        liquid_class_record
    )

    new_liquid_class_record = liquid_class_record.model_copy(deep=True)
    new_liquid_class_record.aspirate.aspiratePosition.offset.x += (
        123  # make it different
    )
    params = LoadLiquidClassParams(
        liquidClassId="liquid-class-1", liquidClassRecord=new_liquid_class_record
    )
    with pytest.raises(LiquidClassRedefinitionError):
        await subject.execute(params)



"""Tests for opentrons.protocol_api.core.engine.ModuleCore."""

import inspect

import pytest
from decoy import Decoy

from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import AbstractModule
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api import validation as mock_validation
from opentrons.protocol_api.core.engine.module_core import ModuleCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_engine import DeckSlotLocation
from opentrons.protocol_engine.clients import SyncClient as EngineClient
from opentrons.protocol_engine.types import ModuleDefinition, ModuleModel
from opentrons.protocols.api_support.types import APIVersion
from opentrons.types import DeckSlotName


@pytest.fixture
def api_version() -> APIVersion:
    """Get mocked api_version."""
    return MAX_SUPPORTED_VERSION


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SynchronousAdapter[AbstractModule]:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SynchronousAdapter[AbstractModule]")  # type: ignore[no-any-return]


@pytest.fixture(autouse=True)
def _mock_validation_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out opentrons.legacy.validation functions."""
    for name, func in inspect.getmembers(mock_validation, inspect.isfunction):
        monkeypatch.setattr(mock_validation, name, decoy.mock(func=func))


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    mock_engine_client: EngineClient,
    api_version: APIVersion,
    mock_sync_module_hardware: SynchronousAdapter[AbstractModule],
    mock_protocol_core: ProtocolCore,
) -> ModuleCore:
    """Get a ModuleCore test subject."""
    return ModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=api_version,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


@pytest.mark.parametrize("api_version", [APIVersion(2, 3)])
def test_api_version(subject: ModuleCore, api_version: APIVersion) -> None:
    """Should return the api_version property."""
    assert subject.api_version == api_version


def test_get_deck_slot(
    decoy: Decoy, subject: ModuleCore, mock_engine_client: EngineClient
) -> None:
    """Should return the deck slot associated to the module id."""
    decoy.when(mock_engine_client.state.modules.get_location("1234")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )

    assert subject.get_deck_slot() == DeckSlotName.SLOT_1


def test_get_deck_slot_id(
    decoy: Decoy, subject: ModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return the correct deck slot string associated with the module and robot type."""
    decoy.when(mock_engine_client.state.modules.get_location("1234")).then_return(
        DeckSlotLocation(slotName=DeckSlotName.SLOT_1)
    )

    decoy.when(mock_engine_client.state.config.robot_type).then_return("OT-3 Standard")

    decoy.when(
        mock_validation.internal_slot_to_public_string(
            DeckSlotName.SLOT_1, "OT-3 Standard"
        )
    ).then_return("foo")

    assert subject.get_deck_slot_id() == "foo"


def test_get_model(
    decoy: Decoy, subject: ModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return the module model."""
    decoy.when(
        mock_engine_client.state.modules.get_connected_model("1234")
    ).then_return(ModuleModel.HEATER_SHAKER_MODULE_V1)

    result = subject.get_model()

    assert result == "heaterShakerModuleV1"


def test_get_display_name(
    decoy: Decoy, subject: ModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return the module display name."""
    module_definition = ModuleDefinition.model_construct(  # type: ignore[call-arg]
        displayName="abra kadabra",
    )
    decoy.when(mock_engine_client.state.modules.get_definition("1234")).then_return(
        module_definition
    )

    assert subject.get_display_name() == "abra kadabra"



"""Test load labware commands."""

import inspect

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.reload_labware import (
    ReloadLabwareImplementation,
    ReloadLabwareParams,
    ReloadLabwareResult,
)
from opentrons.protocol_engine.errors import (
    LabwareNotLoadedError,
)
from opentrons.protocol_engine.execution import EquipmentHandler, ReloadedLabwareData
from opentrons.protocol_engine.resources import labware_validation
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.state.update_types import (
    LabwareLocationUpdate,
    StateUpdate,
)
from opentrons.protocol_engine.types import (
    DeckSlotLocation,
    OnAddressableAreaLocationSequenceComponent,
)
from opentrons.types import DeckSlotName


@pytest.fixture(autouse=True)
def patch_mock_labware_validation(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out labware_validation.py functions."""
    for name, func in inspect.getmembers(labware_validation, inspect.isfunction):
        monkeypatch.setattr(labware_validation, name, decoy.mock(func=func))


async def test_reload_labware_implementation(
    decoy: Decoy,
    well_plate_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A ReloadLabware command should have an execution implementation."""
    subject = ReloadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = ReloadLabwareParams(
        labwareId="my-labware-id",
    )

    decoy.when(
        await equipment.reload_labware(
            labware_id="my-labware-id",
        )
    ).then_return(
        ReloadedLabwareData(
            location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
            offsetId="labware-offset-id",
        )
    )
    decoy.when(
        state_view.geometry.get_predicted_location_sequence(
            DeckSlotLocation(slotName=DeckSlotName.SLOT_4)
        )
    ).then_return([OnAddressableAreaLocationSequenceComponent(addressableAreaName="4")])

    result = await subject.execute(data)

    assert result == SuccessData(
        public=ReloadLabwareResult(
            labwareId="my-labware-id",
            offsetId="labware-offset-id",
            locationSequence=[
                OnAddressableAreaLocationSequenceComponent(addressableAreaName="4")
            ],
        ),
        state_update=StateUpdate(
            labware_location=LabwareLocationUpdate(
                labware_id="my-labware-id",
                new_location=DeckSlotLocation(slotName=DeckSlotName.SLOT_4),
                offset_id="labware-offset-id",
            )
        ),
    )


async def test_reload_labware_raises_labware_does_not_exist(
    decoy: Decoy,
    well_plate_def: LabwareDefinition,
    equipment: EquipmentHandler,
    state_view: StateView,
) -> None:
    """A ReloadLabware command should raise if the specified labware is not loaded."""
    subject = ReloadLabwareImplementation(equipment=equipment, state_view=state_view)

    data = ReloadLabwareParams(
        labwareId="my-labware-id",
    )

    decoy.when(
        await equipment.reload_labware(
            labware_id="my-labware-id",
        )
    ).then_raise(LabwareNotLoadedError("What labware is this!"))

    with pytest.raises(LabwareNotLoadedError):
        await subject.execute(data)



"""Test pick up tip commands."""

from datetime import datetime
from unittest.mock import sentinel

from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import StallOrCollisionDetectedError

from opentrons.protocol_engine import (
    DeckPoint,
    PickUpTipWellLocation,
    WellLocation,
    WellOffset,
)
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pick_up_tip import (
    PickUpTipImplementation,
    PickUpTipParams,
    PickUpTipResult,
    TipPhysicallyMissingError,
)
from opentrons.protocol_engine.errors import PickUpTipTipNotAttachedError
from opentrons.protocol_engine.execution import MovementHandler, TipHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId, TipGeometry, TipRackWellState
from opentrons.types import MountType, Point


async def test_success(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
) -> None:
    """A PickUpTip command should have an execution implementation."""
    subject = PickUpTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
    )

    decoy.when(state_view.pipettes.get_mount("pipette-id")).then_return(MountType.LEFT)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset(x=1, y=2, z=3))
        )
    ).then_return(WellLocation(offset=WellOffset(x=1, y=2, z=3)))

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="A3",
            well_location=WellLocation(offset=WellOffset(x=1, y=2, z=3)),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))

    decoy.when(
        await tip_handler.pick_up_tip(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="A3",
        )
    ).then_return(TipGeometry(length=42, diameter=5, volume=300))

    decoy.when(state_view.pipettes.get_nozzle_configuration("pipette-id")).then_return(
        sentinel.nozzle_configuration
    )
    decoy.when(
        state_view.tips.compute_tips_to_mark_as_used_or_empty(
            "labware-id", "A3", sentinel.nozzle_configuration
        )
    ).then_return(sentinel.tips_to_mark_as_used)

    result = await subject.execute(
        PickUpTipParams(
            pipetteId="pipette-id",
            labwareId="labware-id",
            wellName="A3",
            wellLocation=PickUpTipWellLocation(offset=WellOffset(x=1, y=2, z=3)),
        )
    )

    assert result == SuccessData(
        public=PickUpTipResult(
            tipLength=42,
            tipVolume=300,
            tipDiameter=5,
            position=DeckPoint(x=111, y=222, z=333),
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(labware_id="labware-id", well_name="A3"),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="pipette-id",
                tip_geometry=TipGeometry(length=42, diameter=5, volume=300),
                tip_source=LabwareWellId(labware_id="labware-id", well_name="A3"),
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.EMPTY,
                labware_id="labware-id",
                well_names=sentinel.tips_to_mark_as_used,
            ),
            pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                pipette_id="pipette-id", clean_tip=True
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id", ready_to_aspirate=True
            ),
        ),
    )


async def test_tip_physically_missing_error(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
) -> None:
    """It should return a TipPhysicallyMissingError if the HW API indicates that."""
    subject = PickUpTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
    )

    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    error_id = "error-id"
    error_created_at = datetime(1234, 5, 6)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset())
        )
    ).then_return(WellLocation(offset=WellOffset()))

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(offset=WellOffset()),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_return(Point(x=111, y=222, z=333))
    decoy.when(
        await tip_handler.pick_up_tip(
            pipette_id=pipette_id, labware_id=labware_id, well_name=well_name
        )
    ).then_raise(PickUpTipTipNotAttachedError(tip_geometry=sentinel.tip_geometry))
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_created_at)

    decoy.when(state_view.pipettes.get_nozzle_configuration(pipette_id)).then_return(
        sentinel.nozzle_configuration
    )
    decoy.when(
        state_view.tips.compute_tips_to_mark_as_used_or_empty(
            labware_id, well_name, sentinel.nozzle_configuration
        )
    ).then_return(sentinel.tips_to_mark_as_used)

    result = await subject.execute(
        PickUpTipParams(pipetteId=pipette_id, labwareId=labware_id, wellName=well_name)
    )

    assert result == DefinedErrorData(
        public=TipPhysicallyMissingError.model_construct(
            id=error_id, createdAt=error_created_at, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id", well_name="well-name"
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.EMPTY,
                labware_id="labware-id",
                well_names=sentinel.tips_to_mark_as_used,
            ),
            pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                pipette_id="pipette-id"
            ),
        ),
        state_update_if_false_positive=update_types.StateUpdate(
            pipette_tip_state=update_types.PipetteTipStateUpdate(
                pipette_id="pipette-id",
                tip_geometry=sentinel.tip_geometry,
                tip_source=LabwareWellId(
                    labware_id="labware-id", well_name="well-name"
                ),
            ),
            pipette_aspirated_fluid=update_types.PipetteEmptyFluidUpdate(
                pipette_id="pipette-id", clean_tip=True
            ),
            tips_state=update_types.TipsStateUpdate(
                tip_state=TipRackWellState.EMPTY,
                labware_id="labware-id",
                well_names=sentinel.tips_to_mark_as_used,
            ),
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="pipette-id",
                new_location=LabwareWellId(
                    labware_id="labware-id", well_name="well-name"
                ),
                new_deck_point=DeckPoint(x=111, y=222, z=333),
            ),
        ),
    )


async def test_stall_error(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    tip_handler: TipHandler,
    model_utils: ModelUtils,
) -> None:
    """It should return a TipPhysicallyMissingError if the HW API indicates that."""
    subject = PickUpTipImplementation(
        state_view=state_view,
        movement=movement,
        tip_handler=tip_handler,
        model_utils=model_utils,
    )

    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    error_id = "error-id"
    error_created_at = datetime(1234, 5, 6)

    decoy.when(
        state_view.geometry.convert_pick_up_tip_well_location(
            well_location=PickUpTipWellLocation(offset=WellOffset())
        )
    ).then_return(WellLocation(offset=WellOffset()))

    decoy.when(
        await movement.move_to_well(
            pipette_id="pipette-id",
            labware_id="labware-id",
            well_name="well-name",
            well_location=WellLocation(offset=WellOffset()),
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        )
    ).then_raise(StallOrCollisionDetectedError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_created_at)

    result = await subject.execute(
        PickUpTipParams(pipetteId=pipette_id, labwareId=labware_id, wellName=well_name)
    )

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id, createdAt=error_created_at, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.CLEAR,
        ),
    )



from opentrons.protocol_api.protocol_context import ProtocolContext

metadata = {"author": "MISTER FIXTURE", "apiLevel": "2.0"}


def run(protocol_context: ProtocolContext) -> None:
    tip_rack = protocol_context.load_labware("opentrons_96_tiprack_10ul", "3")
    plate = protocol_context.load_labware(
        "custom_labware", "1", namespace="custom_beta"
    )

    pipette = protocol_context.load_instrument(
        "p10_single", "left", tip_racks=[tip_rack]
    )

    csv_data = protocol_context.bundled_data["data.txt"].decode("utf-8")

    for volume in csv_data.split(","):
        v = float(volume.strip())
        pipette.transfer(v, plate.wells("A1"), plate.wells("A4"))



import logging
from dataclasses import dataclass
from datetime import datetime
from functools import lru_cache
from typing import Any, List, Optional, cast

import numpy as np
from numpy.typing import NDArray

from .util import DeckTransformState
from opentrons import config
from opentrons.calibration_storage import (
    get_robot_deck_attitude,
    save_robot_deck_attitude,
    types,
)
from opentrons.config.robot_configs import (
    default_ot2_deck_calibration,
    get_legacy_gantry_calibration,
)
from opentrons.util import linal

log = logging.getLogger(__name__)


@dataclass
class DeckCalibration:
    attitude: types.AttitudeMatrix
    source: types.SourceType
    status: types.CalibrationStatus
    belt_attitude: Optional[types.AttitudeMatrix] = None
    last_modified: Optional[datetime] = None
    pipette_calibrated_with: Optional[str] = None
    tiprack: Optional[str] = None


@dataclass
class RobotCalibration:
    deck_calibration: DeckCalibration


def validate_attitude_deck_calibration(
    deck_cal: DeckCalibration,
) -> DeckTransformState:
    """
    This function determines whether the deck calibration is valid
    or not based on the following use-cases:

    TODO(lc, 8/10/2020): As with the OT2, expand on this method, or create
    another method to diagnose bad instrument offset data
    """
    curr_cal: linal.DoubleMatrix = np.array(deck_cal.attitude)
    row, _ = curr_cal.shape
    rank: int = np.linalg.matrix_rank(curr_cal)
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif not deck_cal.last_modified:
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def validate_gantry_calibration(gantry_cal: List[List[float]]) -> DeckTransformState:
    """
    This function determines whether the gantry calibration is valid
    or not based on the following use-cases:
    """
    curr_cal: linal.DoubleMatrix = np.array(gantry_cal)
    row, _ = curr_cal.shape

    rank: int = np.linalg.matrix_rank(curr_cal)

    id_matrix = linal.identity_deck_transform()

    z = abs(curr_cal[2][-1])

    outofrange = z < 16 or z > 34
    if row != rank:
        # Check that the matrix is non-singular
        return DeckTransformState.SINGULARITY
    elif np.array_equal(curr_cal, id_matrix):
        # Check that the matrix is not an identity
        return DeckTransformState.IDENTITY
    elif outofrange:
        # Check that the matrix is not out of range.
        return DeckTransformState.BAD_CALIBRATION
    else:
        # Transform as it stands is sufficient.
        return DeckTransformState.OK


def migrate_affine_xy_to_attitude(
    gantry_cal: List[List[float]],
) -> types.AttitudeMatrix:
    masked_transform: NDArray[np.bool_] = np.array(
        [
            [True, True, True, False],
            [True, True, True, False],
            [False, False, False, False],
            [False, False, False, False],
        ]
    )
    masked_array: np.ma.MaskedArray[Any, np.dtype[np.float64]] = np.ma.masked_array(  # type: ignore
        gantry_cal, ~masked_transform
    )
    attitude_array: linal.DoubleMatrix = np.zeros((3, 3))
    np.put(attitude_array, [0, 1, 2], masked_array[0].compressed())
    np.put(attitude_array, [3, 4, 5], masked_array[1].compressed())
    np.put(attitude_array, 8, 1)
    return cast(List[List[float]], attitude_array.tolist())


def save_attitude_matrix(
    expected: linal.SolvePoints,
    actual: linal.SolvePoints,
    pipette_id: str,
    tiprack_hash: str,
) -> None:
    attitude = linal.solve_attitude(expected, actual)
    save_robot_deck_attitude(
        attitude,
        pipette_id,
        tiprack_hash,
    )


def load_attitude_matrix() -> DeckCalibration:
    calibration_data = get_robot_deck_attitude()
    gantry_cal = get_legacy_gantry_calibration()
    if not calibration_data and gantry_cal:
        if validate_gantry_calibration(gantry_cal) == DeckTransformState.OK:
            log.debug(
                "Attitude deck calibration matrix not found. Migrating "
                "existing affine deck calibration matrix to {}".format(
                    config.get_opentrons_path("robot_calibration_dir")
                )
            )
            attitude = migrate_affine_xy_to_attitude(gantry_cal)
            save_robot_deck_attitude(
                transform=attitude,
                pip_id=None,
                lw_hash=None,
                source=types.SourceType.legacy,
            )
            calibration_data = get_robot_deck_attitude()

    if calibration_data:
        return DeckCalibration(
            attitude=calibration_data.attitude,
            source=calibration_data.source,
            status=types.CalibrationStatus(**calibration_data.status.model_dump()),
            last_modified=calibration_data.last_modified,
            pipette_calibrated_with=calibration_data.pipette_calibrated_with,
            tiprack=calibration_data.tiprack,
        )
    else:
        # load default if deck calibration data do not exist
        return DeckCalibration(
            attitude=default_ot2_deck_calibration(),
            source=types.SourceType.default,
            status=types.CalibrationStatus(),
        )


def load() -> RobotCalibration:
    return RobotCalibration(deck_calibration=load_attitude_matrix())


class RobotCalibrationProvider:
    def __init__(self) -> None:
        self._robot_calibration = load()

    @lru_cache(1)
    def _validate(self) -> DeckTransformState:
        return validate_attitude_deck_calibration(
            self._robot_calibration.deck_calibration
        )

    @property
    def robot_calibration(self) -> RobotCalibration:
        return self._robot_calibration

    def reset_robot_calibration(self) -> None:
        self._validate.cache_clear()
        self._robot_calibration = load()

    def reset_deck_calibration(self) -> None:
        self._robot_calibration = load()

    def load_deck_calibration(self) -> None:
        self._robot_calibration = load()

    def set_robot_calibration(self, robot_calibration: RobotCalibration) -> None:
        self._validate.cache_clear()
        self._robot_calibration = robot_calibration

    def validate_calibration(self) -> DeckTransformState:
        """
        The lru cache decorator is currently not supported by the
        ThreadManager. To work around this, we need to wrap the
        actual function around a dummy outer function.

        Once decorators are more fully supported, we can remove this.
        """
        return self._validate()

    def build_temporary_identity_calibration(self) -> RobotCalibration:
        """
        Get a temporary identity deck cal suitable for use during
        calibration
        """
        return RobotCalibration(
            deck_calibration=DeckCalibration(
                attitude=default_ot2_deck_calibration(),
                source=types.SourceType.default,
                status=types.CalibrationStatus(),
            )
        )
